# standard libraries
import contextlib
import copy
import logging
import random
import threading
import unittest

# third party libraries
import numpy
import scipy

# local libraries
from nion.data import Calibration
from nion.data import Core
from nion.data import Image
from nion.swift import Application
from nion.swift import DocumentController
from nion.swift import Facade
from nion.swift.model import DataItem
from nion.swift.model import DocumentModel
from nion.swift.model import Graphics
from nion.swift.model import Symbolic
from nion.swift.model import Utility
from nion.ui import TestUI
from nion.utils import Geometry


Facade.initialize()


class TestSymbolicClass(unittest.TestCase):

    def setUp(self):
        self.app = Application.Application(TestUI.UserInterface(), set_global=True)

    def tearDown(self):
        pass

    def test_unary_inversion_returns_inverted_data(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            d = numpy.zeros((8, 8), dtype=numpy.uint32)
            d[:] = random.randint(1, 100)
            data_item = DataItem.DataItem(d)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("-a.xdata"))
            computation.create_object("a", document_model.get_object_specifier(data_item))
            data = DocumentModel.evaluate_data(computation).data
            assert numpy.array_equal(data, -d)

    def test_binary_addition_returns_added_data(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            d1 = numpy.zeros((8, 8), dtype=numpy.uint32)
            d1[:] = random.randint(1, 100)
            data_item1 = DataItem.DataItem(d1)
            d2 = numpy.zeros((8, 8), dtype=numpy.uint32)
            d2[:] = random.randint(1, 100)
            data_item2 = DataItem.DataItem(d2)
            document_model.append_data_item(data_item1)
            document_model.append_data_item(data_item2)
            computation = document_model.create_computation(Symbolic.xdata_expression("a.xdata + b.xdata"))
            computation.create_object("a", document_model.get_object_specifier(data_item1))
            computation.create_object("b", document_model.get_object_specifier(data_item2))
            data = DocumentModel.evaluate_data(computation).data
            assert numpy.array_equal(data, d1 + d2)

    def test_binary_multiplication_with_scalar_returns_multiplied_data(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            d = numpy.zeros((8, 8), dtype=numpy.uint32)
            d[:] = random.randint(1, 100)
            data_item = DataItem.DataItem(d)
            document_model.append_data_item(data_item)
            computation1 = document_model.create_computation(Symbolic.xdata_expression("a.xdata * 5"))
            computation1.create_object("a", document_model.get_object_specifier(data_item))
            data1 = DocumentModel.evaluate_data(computation1).data
            computation2 = document_model.create_computation(Symbolic.xdata_expression("5 * a.xdata"))
            computation2.create_object("a", document_model.get_object_specifier(data_item))
            data2 = DocumentModel.evaluate_data(computation2).data
            assert numpy.array_equal(data1, d * 5)
            assert numpy.array_equal(data2, d * 5)

    def test_subtract_min_returns_subtracted_min(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            d = numpy.zeros((8, 8), dtype=numpy.uint32)
            d[:] = random.randint(1, 100)
            data_item = DataItem.DataItem(d)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("a.xdata - numpy.amin(a.data)"))
            computation.create_object("a", document_model.get_object_specifier(data_item))
            data = DocumentModel.evaluate_data(computation).data
            assert numpy.array_equal(data, d - numpy.amin(d))

    def test_ability_to_take_slice(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            d = numpy.zeros((4, 8, 8), dtype=numpy.uint32)
            d[:] = random.randint(1, 100)
            data_item = DataItem.DataItem(d)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("a.xdata[:,4,4]"))
            computation.create_object("a", document_model.get_object_specifier(data_item))
            data = DocumentModel.evaluate_data(computation).data
            assert numpy.array_equal(data, d[:,4,4])

    def test_ability_to_take_slice_on_1d_data(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            d = numpy.random.randn(8)
            data_item = DataItem.DataItem(d)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("a.xdata[2:6]"))
            computation.create_object("a", document_model.get_object_specifier(data_item))
            data = DocumentModel.evaluate_data(computation).data
            assert numpy.array_equal(data, d[2:6])

    def test_slice_with_empty_dimension_produces_error(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            d = numpy.zeros((4, 8, 8), dtype=numpy.uint32)
            d[:] = random.randint(1, 100)
            data_item = DataItem.DataItem(d)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("a.xdata[2:2, :, :]"))
            computation.create_object("a", document_model.get_object_specifier(data_item))
            self.assertIsNone(DocumentModel.evaluate_data(computation))

    def test_ability_to_take_slice_with_ellipses_produces_correct_data(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            d = numpy.zeros((4, 8, 8), dtype=numpy.uint32)
            d[:] = random.randint(1, 100)
            data_item = DataItem.DataItem(d)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("a.xdata[2, ...]"))
            computation.create_object("a", document_model.get_object_specifier(data_item))
            data = DocumentModel.evaluate_data(computation).data
            assert numpy.array_equal(data, d[2, ...])

    def test_ability_to_take_slice_with_ellipses_produces_correct_calibration(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            d = numpy.zeros((4, 8, 8), dtype=numpy.uint32)
            d[:] = random.randint(1, 100)
            data_item = DataItem.DataItem(d)
            data_item.set_dimensional_calibrations([Calibration.Calibration(10, 20, "m"), Calibration.Calibration(11, 21, "mm"), Calibration.Calibration(12, 22, "nm")])
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("a.xdata[2, ...]"))
            computation.create_object("a", document_model.get_object_specifier(data_item))
            data_and_metadata = DocumentModel.evaluate_data(computation)
            self.assertEqual(len(data_and_metadata.data_shape), len(data_and_metadata.dimensional_calibrations))
            self.assertEqual("mm", data_and_metadata.dimensional_calibrations[0].units)
            self.assertEqual("nm", data_and_metadata.dimensional_calibrations[1].units)

    def test_ability_to_take_slice_with_newaxis(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            d = numpy.zeros((8, 8), dtype=numpy.uint32)
            d[:] = random.randint(1, 100)
            data_item = DataItem.DataItem(d)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("a.xdata[numpy.newaxis, ...]"))
            computation.create_object("a", document_model.get_object_specifier(data_item))
            data = DocumentModel.evaluate_data(computation).data
            assert numpy.array_equal(data, d[numpy.newaxis, ...])

    def test_ability_to_take_1d_slice_with_newaxis(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            d = numpy.zeros((8,), dtype=numpy.uint32)
            d[:] = random.randint(1, 100)
            data_item = DataItem.DataItem(d)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("a.xdata[..., numpy.newaxis]"))
            computation.create_object("a", document_model.get_object_specifier(data_item))
            data = DocumentModel.evaluate_data(computation).data
            assert numpy.array_equal(data, d[..., numpy.newaxis])

    def test_slice_sum_sums_correct_slices(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            d = numpy.random.randn(4, 4, 16)
            data_item = DataItem.DataItem(d)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("xd.slice_sum(a.xdata, 4, 6)"))
            computation.create_object("a", document_model.get_object_specifier(data_item))
            data = DocumentModel.evaluate_data(computation).data
            assert numpy.array_equal(data, numpy.sum(d[..., 1:7], -1))

    def test_reshape_1d_to_2d_produces_correct_data(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            d = numpy.random.randn(4)
            data_item = DataItem.DataItem(d)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("xd.reshape(a.xdata, (2, 2))"))
            computation.create_object("a", document_model.get_object_specifier(data_item))
            data = DocumentModel.evaluate_data(computation).data
            assert numpy.array_equal(data, numpy.reshape(d, (2, 2)))
            computation = document_model.create_computation(Symbolic.xdata_expression("xd.reshape(a.xdata, (4, -1))"))
            computation.create_object("a", document_model.get_object_specifier(data_item))
            data = DocumentModel.evaluate_data(computation).data
            assert numpy.array_equal(data, numpy.reshape(d, (4, -1)))
            computation = document_model.create_computation(Symbolic.xdata_expression("xd.reshape(a.xdata, (-1, 4))"))
            computation.create_object("a", document_model.get_object_specifier(data_item))
            data = DocumentModel.evaluate_data(computation).data
            assert numpy.array_equal(data, numpy.reshape(d, (-1, 4)))

    def test_reshape_1d_to_2d_preserves_calibration(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            d = numpy.random.randn(4)
            data_item = DataItem.DataItem(d)
            data_item.set_dimensional_calibrations([Calibration.Calibration(1.1, 2.1, "m")])
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("xd.reshape(a.xdata, (4, -1))"))
            computation.create_object("a", document_model.get_object_specifier(data_item))
            data_and_metadata = DocumentModel.evaluate_data(computation)
            self.assertEqual("m", data_and_metadata.dimensional_calibrations[0].units)
            self.assertEqual("", data_and_metadata.dimensional_calibrations[1].units)
            computation = document_model.create_computation(Symbolic.xdata_expression("xd.reshape(a.xdata, (-1, 4))"))
            computation.create_object("a", document_model.get_object_specifier(data_item))
            data_and_metadata = DocumentModel.evaluate_data(computation)
            self.assertEqual("", data_and_metadata.dimensional_calibrations[0].units)
            self.assertEqual("m", data_and_metadata.dimensional_calibrations[1].units)

    def test_reshape_2d_n_x_1_to_1d_preserves_calibration(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            d = numpy.random.randn(4, 1)
            data_item = DataItem.DataItem(d)
            data_item.set_dimensional_calibrations([Calibration.Calibration(1.1, 2.1, "m"), Calibration.Calibration()])
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("xd.reshape(a.xdata, (4, ))"))
            computation.create_object("a", document_model.get_object_specifier(data_item))
            data_and_metadata = DocumentModel.evaluate_data(computation)
            self.assertEqual(1, len(data_and_metadata.dimensional_calibrations))
            self.assertEqual("m", data_and_metadata.dimensional_calibrations[0].units)

    def test_reshape_to_match_another_item(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            d = numpy.random.randn(4)
            d2 = numpy.random.randn(2, 2)
            data_item = DataItem.DataItem(d)
            document_model.append_data_item(data_item)
            data_item2 = DataItem.DataItem(d2)
            document_model.append_data_item(data_item2)
            computation = document_model.create_computation(Symbolic.xdata_expression("xd.reshape(a.xdata, b.xdata.data_shape)"))
            computation.create_object("a", document_model.get_object_specifier(data_item))
            computation.create_object("b", document_model.get_object_specifier(data_item2))
            data = DocumentModel.evaluate_data(computation).data
            assert numpy.array_equal(data, numpy.reshape(d, (2, 2)))

    def test_concatenate_two_images(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            d = numpy.random.randn(4, 4)
            data_item = DataItem.DataItem(d)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("xd.concatenate((a.xdata[0:2, 0:2], a.xdata[2:4, 2:4]))"))
            computation.create_object("a", document_model.get_object_specifier(data_item))
            data = DocumentModel.evaluate_data(computation).data
            assert numpy.array_equal(data, numpy.concatenate((d[0:2, 0:2], d[2:4, 2:4])))

    def test_concatenate_keeps_calibrations_in_non_axis_dimensions(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            d = numpy.random.randn(4, 4)
            data_item = DataItem.DataItem(d)
            data_item.set_intensity_calibration(Calibration.Calibration(1.0, 2.0, "nm"))
            data_item.set_dimensional_calibrations([Calibration.Calibration(1.1, 2.1, "m"), Calibration.Calibration(1.2, 2.2, "s")])
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("xd.concatenate((a.xdata[0:2, 0:2], a.xdata[2:4, 2:4]))"))
            computation.create_object("a", document_model.get_object_specifier(data_item))
            data_and_metadata = DocumentModel.evaluate_data(computation)
            self.assertEqual("nm", data_and_metadata.intensity_calibration.units)
            self.assertEqual("m", data_and_metadata.dimensional_calibrations[0].units)
            self.assertEqual("", data_and_metadata.dimensional_calibrations[1].units)

    def test_concatenate_along_alternate_axis_images(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            d = numpy.random.randn(4)
            data_item = DataItem.DataItem(d)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("xd.concatenate((xd.reshape(a.xdata, (1, -1)), xd.reshape(a.xdata, (1, -1))), 0)"))
            computation.create_object("a", document_model.get_object_specifier(data_item))
            data = DocumentModel.evaluate_data(computation).data
            assert numpy.array_equal(data, numpy.concatenate((numpy.reshape(d, (1, -1)), numpy.reshape(d, (1, -1))), 0))

    def test_concatenate_three_images_along_second_axis(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            d = numpy.random.randn(4, 4)
            data_item = DataItem.DataItem(d)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("xd.concatenate((a.xdata[0:2, 0:2], a.xdata[1:3, 1:3], a.xdata[2:4, 2:4]), 1)"))
            computation.create_object("a", document_model.get_object_specifier(data_item))
            data = DocumentModel.evaluate_data(computation).data
            assert numpy.array_equal(data, numpy.concatenate((d[0:2, 0:2], d[1:3, 1:3], d[2:4, 2:4]), 1))

    def test_ability_to_write_read_basic_nodes(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            src_data = numpy.zeros((8, 8), dtype=numpy.uint32)
            src_data[:] = random.randint(1, 100)
            data_item = DataItem.DataItem(src_data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("-a.xdata / numpy.average(a.data) * 5"))
            computation.create_object("a", document_model.get_object_specifier(data_item))
            data_node_dict = computation.write_to_dict()
            computation2 = document_model.create_computation()
            computation2.read_from_dict(data_node_dict)
            computation2.needs_update = True
            computation2.bind(document_model)
            data = DocumentModel.evaluate_data(computation).data
            data2 = DocumentModel.evaluate_data(computation2).data
            assert numpy.array_equal(data, -src_data / numpy.average(src_data) * 5)
            assert numpy.array_equal(data, data2)

    def test_make_operation_works_without_exception_and_produces_correct_data(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            d = numpy.zeros((8, 8), dtype=numpy.uint32)
            d[:] = random.randint(1, 100)
            data_item = DataItem.DataItem(d)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("-a.xdata / numpy.average(a.data) * 5"))
            computation.create_object("a", document_model.get_object_specifier(data_item))
            data = DocumentModel.evaluate_data(computation).data
            assert numpy.array_equal(data, -d / numpy.average(d) * 5)

    def test_fft_returns_complex_data(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            d = numpy.random.randn(64, 64)
            data_item = DataItem.DataItem(d)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("xd.fft(a.xdata)"))
            computation.create_object("a", document_model.get_object_specifier(data_item))
            data = DocumentModel.evaluate_data(computation).data
            assert numpy.array_equal(data, scipy.fftpack.fftshift(numpy.fft.fft2(d) * 1.0 / numpy.sqrt(d.shape[1] * d.shape[0])))

    def test_gaussian_blur_handles_scalar_argument(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            d = numpy.random.randn(64, 64)
            data_item = DataItem.DataItem(d)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("xd.gaussian_blur(a.xdata, 4.0)"))
            computation.create_object("a", document_model.get_object_specifier(data_item))
            data = DocumentModel.evaluate_data(computation).data
            assert numpy.array_equal(data, scipy.ndimage.gaussian_filter(d, sigma=4.0))

    def test_transpose_flip_handles_args(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            d = numpy.random.randn(30, 60)
            data_item = DataItem.DataItem(d)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("xd.transpose_flip(a.xdata, flip_v=True)"))
            computation.create_object("a", document_model.get_object_specifier(data_item))
            data = DocumentModel.evaluate_data(computation).data
            assert numpy.array_equal(data, numpy.flipud(d))

    def test_crop_handles_args(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            d = numpy.random.randn(64, 64)
            data_item = DataItem.DataItem(d)
            region = Graphics.RectangleGraphic()
            region.center = 0.41, 0.51
            region.size = 0.52, 0.42
            data_item.displays[0].add_graphic(region)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("xd.crop(a.xdata, regionA.bounds)"))
            computation.create_object("a", document_model.get_object_specifier(data_item))
            computation.create_object("regionA", document_model.get_object_specifier(region))
            data = DocumentModel.evaluate_data(computation).data
            assert numpy.array_equal(data, d[9:42, 19:45])

    def test_evaluate_computation_gives_correct_value(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            data = numpy.ones((2, 2), numpy.double)
            data_item = DataItem.DataItem(data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("-a.xdata"))
            computation.create_object("a", document_model.get_object_specifier(data_item))
            data_and_metadata = DocumentModel.evaluate_data(computation)
            self.assertTrue(numpy.array_equal(data_and_metadata.data, -data))

    def test_computation_fires_needs_update_event_when_data_changes(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            data = numpy.ones((2, 2), numpy.double)
            data_item = DataItem.DataItem(data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("-a.xdata"))
            computation.create_object("a", document_model.get_object_specifier(data_item))
            needs_update_ref = [False]
            def needs_update():
                needs_update_ref[0] = True
            needs_update_event_listener = computation.computation_mutated_event.listen(needs_update)
            with contextlib.closing(needs_update_event_listener):
                with data_item.data_ref() as dr:
                    dr.data += 1.5
            self.assertTrue(needs_update_ref[0])

    def test_computation_fires_needs_update_event_when_display_data_changes(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            data = ((numpy.random.randn(2, 2, 2, 2) + 1) * 10).astype(numpy.uint32)
            data_item = DataItem.DataItem(data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("a.display_xdata"))
            computation.create_object("a", document_model.get_object_specifier(data_item))
            computed_data_item = DataItem.DataItem(data.copy())
            document_model.append_data_item(computed_data_item)
            document_model.set_data_item_computation(computed_data_item, computation)
            # verify assumptions
            document_model.recompute_all()
            self.assertTrue(numpy.array_equal(data[0, 0, ...], computed_data_item.data))
            # change display, check
            data_item.displays[0].collection_index = 1, 1
            document_model.recompute_all()
            self.assertTrue(numpy.array_equal(data[1, 1, ...], computed_data_item.data))

    def test_computation_fires_needs_update_event_when_metadata_changes(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            data = numpy.ones((2, 2), numpy.double)
            data_item = DataItem.DataItem(data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("-a.xdata"))
            computation.create_object("a", document_model.get_object_specifier(data_item))
            needs_update_ref = [False]
            def needs_update():
                needs_update_ref[0] = True
            needs_update_event_listener = computation.computation_mutated_event.listen(needs_update)
            with contextlib.closing(needs_update_event_listener):
                metadata = data_item.metadata
                metadata["abc"] = 1
                data_item.metadata = metadata
            self.assertTrue(needs_update_ref[0])

    def test_computation_does_not_update_when_graphic_changes_on_source(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            data = numpy.ones((2, 2), numpy.double)
            data_item = DataItem.DataItem(data)
            data_item.displays[0].add_graphic(Graphics.PointGraphic())
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("-a.xdata"))
            computation.create_object("a", document_model.get_object_specifier(data_item))
            needs_update_ref = [False]
            def needs_update():
                needs_update_ref[0] = True
            needs_update_event_listener = computation.computation_mutated_event.listen(needs_update)
            with contextlib.closing(needs_update_event_listener):
                data_item.displays[0].graphics[0].position = (0.3, 0.4)
            self.assertFalse(needs_update_ref[0])

    def test_computation_fires_needs_update_event_when_object_property(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            data = numpy.random.randn(64, 64)
            data_item = DataItem.DataItem(data)
            region = Graphics.RectangleGraphic()
            region.center = 0.41, 0.51
            region.size = 0.52, 0.42
            data_item.displays[0].add_graphic(region)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("xd.crop(a.xdata, regionA.bounds)"))
            computation.create_object("a", document_model.get_object_specifier(data_item))
            computation.create_object("regionA", document_model.get_object_specifier(region))
            needs_update_ref = [False]
            def needs_update():
                needs_update_ref[0] = True
            needs_update_event_listener = computation.computation_mutated_event.listen(needs_update)
            with contextlib.closing(needs_update_event_listener):
                data_item.displays[0].graphics[0].size = 0.53, 0.43
            self.assertTrue(needs_update_ref[0])

    def test_computation_fires_needs_update_event_when_variable_or_object_added(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            data = numpy.random.randn(64, 64)
            data_item = DataItem.DataItem(data)
            region = Graphics.RectangleGraphic()
            region.center = 0.41, 0.51
            region.size = 0.52, 0.42
            data_item.displays[0].add_graphic(region)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("a.xdata + n"))
            needs_update_ref = [False]
            def needs_update():
                needs_update_ref[0] = True
            needs_update_event_listener = computation.computation_mutated_event.listen(needs_update)
            computation.create_object("a", document_model.get_object_specifier(data_item))
            self.assertTrue(needs_update_ref[0])
            needs_update_ref[0] = False
            computation.create_variable("x", value_type="integral", value=5)
            self.assertTrue(needs_update_ref[0])

    def test_computation_handles_data_lookups(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = self.app.create_document_controller(document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            d = numpy.random.randn(2, 2)
            data_item = DataItem.DataItem(d)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("-api.library.get_data_item_by_uuid(uuid.UUID('{}')).xdata".format(str(data_item.uuid))))
            data_and_metadata = DocumentModel.evaluate_data(computation)
            assert numpy.array_equal(data_and_metadata.data, -d)

    def test_computation_handles_region_lookups(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = self.app.create_document_controller(document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            d = numpy.random.randn(100, 100)
            data_item = DataItem.DataItem(d)
            region = Graphics.RectangleGraphic()
            region.center = 0.5, 0.5
            region.size = 0.6, 0.4
            data_item.displays[0].add_graphic(region)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("xd.crop(a.xdata, api.library.get_graphic_by_uuid(uuid.UUID('{}')).bounds)".format(str(region.uuid))))
            computation.create_object("a", document_model.get_object_specifier(data_item))
            data_and_metadata = DocumentModel.evaluate_data(computation)
            assert numpy.array_equal(data_and_metadata.data, d[20:80, 30:70])

    def test_computation_does_not_copy_metadata_during_computation(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            d = numpy.zeros((8, 8), dtype=numpy.uint32)
            d[:] = random.randint(1, 100)
            data_item = DataItem.DataItem(d)
            data_item.metadata = {"abc": 1}
            data_item.set_intensity_calibration(Calibration.Calibration(1.0, 2.0, "nm"))
            data_item.set_dimensional_calibrations([Calibration.Calibration(1.1, 2.1, "m"), Calibration.Calibration(1.2, 2.2, "m")])
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("-a.xdata / numpy.average(a.data) * 5"))
            computation.create_object("a", document_model.get_object_specifier(data_item))
            data_and_metadata = DocumentModel.evaluate_data(computation)
            self.assertEqual(data_and_metadata.metadata, dict())
            self.assertEqual(data_and_metadata.intensity_calibration, data_item.intensity_calibration)
            self.assertEqual(data_and_metadata.dimensional_calibrations, data_item.dimensional_calibrations)

    def test_remove_data_item_with_computation_succeeds(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            d = numpy.ones((8, 8), dtype=numpy.uint32)
            data_item = DataItem.DataItem(d)
            document_model.append_data_item(data_item)
            map = {"a": document_model.get_object_specifier(data_item)}
            new_data_item = document_controller.processing_computation("-a", map)
            document_model.recompute_all()
            document_model.remove_data_item(new_data_item)

    def test_evaluate_corrupt_computation_gives_sensible_response(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            data = numpy.ones((2, 2), numpy.double)
            data_item = DataItem.DataItem(data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("(a.xdata++)"))
            computation.create_object("a", document_model.get_object_specifier(data_item))
            data_and_metadata = DocumentModel.evaluate_data(computation)
            self.assertIsNone(data_and_metadata)

    def test_evaluate_computation_with_invalid_source_gives_sensible_response(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            data = numpy.ones((2, 2), numpy.double)
            data_item = DataItem.DataItem(data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("a.xdata + e"))
            computation.create_object("a", document_model.get_object_specifier(data_item))
            data_and_metadata = DocumentModel.evaluate_data(computation)
            self.assertIsNone(data_and_metadata)

    def test_evaluate_computation_with_invalid_function_in_document_fails_cleanly(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            data = numpy.ones((2, 2), numpy.double)
            data_item = DataItem.DataItem(data)
            document_model.append_data_item(data_item)
            map = {"a": document_model.get_object_specifier(data_item)}
            document_controller.processing_computation("void(a,2)", map)
            document_model.recompute_all()

    def test_computation_changed_updates_evaluated_data(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            data = numpy.ones((2, 2), numpy.double)
            data_item = DataItem.DataItem(data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("-a.xdata"))
            computation.create_object("a", document_model.get_object_specifier(data_item))
            data_and_metadata = DocumentModel.evaluate_data(computation)
            self.assertTrue(numpy.array_equal(data_and_metadata.data, -data))
            computation.expression = Symbolic.xdata_expression("-2 * a.xdata")
            data_and_metadata = DocumentModel.evaluate_data(computation)
            self.assertTrue(numpy.array_equal(data_and_metadata.data, -data*2))

    def test_changing_computation_updates_data_item(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            data = numpy.random.randn(2, 2)
            data_item = DataItem.DataItem(data)
            document_model.append_data_item(data_item)
            map = {"a": document_model.get_object_specifier(data_item)}
            computed_data_item = document_controller.processing_computation(Symbolic.xdata_expression("-a.xdata"), map)
            document_model.recompute_all()
            self.assertTrue(numpy.array_equal(computed_data_item.data, -data))
            document_model.get_data_item_computation(computed_data_item).expression = Symbolic.xdata_expression("-a.xdata * 2")
            document_model.recompute_all()
            self.assertTrue(numpy.array_equal(computed_data_item.data, -data * 2))

    def test_unary_functions_return_correct_dimensions(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            data = numpy.random.randn(2, 2)
            data_item = DataItem.DataItem(data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.data_expression("numpy.sin(a.data)"))
            computation.create_object("a", document_model.get_object_specifier(data_item))
            data_and_metadata = DocumentModel.evaluate_data(computation)
            self.assertEqual(len(data_and_metadata.dimensional_calibrations), 2)

    def test_computation_stores_original_text(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            data = numpy.random.randn(2, 2)
            data_item = DataItem.DataItem(data)
            document_model.append_data_item(data_item)
            data_expression = Symbolic.data_expression("numpy.sin(a.data)")
            computation = document_model.create_computation(data_expression)
            computation.create_object("a", document_model.get_object_specifier(data_item))
            self.assertEqual(computation.expression, data_expression)

    def test_computation_stores_error_and_original_text(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            data = numpy.random.randn(2, 2)
            data_item = DataItem.DataItem(data)
            document_model.append_data_item(data_item)
            xdata_expression = Symbolic.xdata_expression("xyz(a.xdata)")
            computation = document_model.create_computation(xdata_expression)
            computation.create_object("a", document_model.get_object_specifier(data_item))
            data_and_metadata = DocumentModel.evaluate_data(computation)
            self.assertIsNone(data_and_metadata)
            self.assertTrue(computation.error_text is not None and len(computation.error_text) > 0)
            self.assertEqual(computation.expression, xdata_expression)

    def test_computation_reloads_missing_scalar_function(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            src_data = numpy.zeros((8, 8), dtype=numpy.uint32)
            src_data[:] = random.randint(0, 100)
            data_item = DataItem.DataItem(src_data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("numpy.average(a.data)"))
            computation.create_object("a", document_model.get_object_specifier(data_item))
            data_node_dict = computation.write_to_dict()
            data_node_dict['original_expression'] = "missing(a.xdata)"
            computation2 = document_model.create_computation()
            computation2.read_from_dict(data_node_dict)
            self.assertIsNone(DocumentModel.evaluate_data(computation2))

    def test_computation_can_extract_item_from_scalar_tuple(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            data = ((numpy.random.randn(4, 2) + 1) * 10).astype(numpy.uint32)
            data_item = DataItem.DataItem(data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("a.xdata + a.xdata.data_shape[1] + a.xdata.data_shape[0]"))
            computation.create_object("a", document_model.get_object_specifier(data_item))
            data_and_metadata = DocumentModel.evaluate_data(computation)
            self.assertTrue(numpy.array_equal(data_and_metadata.data, data + 6))

    def test_columns_and_rows_and_radius_functions_return_correct_values(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            data = ((numpy.random.randn(10, 8) + 1) * 10).astype(numpy.uint32)
            data_item = DataItem.DataItem(data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("xd.row(a.xdata.data_shape, -1, 1) + xd.column(a.xdata.data_shape, -1, 1) + xd.radius(a.xdata.data_shape)"))
            computation.create_object("a", document_model.get_object_specifier(data_item))
            data_and_metadata = DocumentModel.evaluate_data(computation)
            icol, irow = numpy.meshgrid(numpy.linspace(-1, 1, 8), numpy.linspace(-1, 1, 10))
            self.assertTrue(numpy.array_equal(data_and_metadata.data, icol + irow + numpy.sqrt(pow(icol, 2) + pow(irow, 2))))

    def test_copying_data_item_with_computation_copies_computation(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            data = ((numpy.random.randn(10, 8) + 1) * 10).astype(numpy.uint32)
            data_item = DataItem.DataItem(data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("-a.xdata"))
            computation.create_object("a", document_model.get_object_specifier(data_item))
            computed_data_item = DataItem.DataItem(data.copy())
            document_model.append_data_item(computed_data_item)
            document_model.set_data_item_computation(computed_data_item, computation)
            document_model.recompute_all()
            copied_data_item = document_model.copy_data_item(computed_data_item)
            self.assertIsNotNone(document_model.get_data_item_computation(copied_data_item))
            self.assertEqual(document_model.get_data_item_computation(computed_data_item).error_text, document_model.get_data_item_computation(copied_data_item).error_text)
            self.assertEqual(document_model.get_data_item_computation(computed_data_item).expression, document_model.get_data_item_computation(copied_data_item).expression)

    def test_changing_computation_source_data_updates_computation(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            data = ((numpy.random.randn(10, 8) + 1) * 10).astype(numpy.uint32)
            data_item = DataItem.DataItem(data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("-a.xdata"))
            computation.create_object("a", document_model.get_object_specifier(data_item))
            computed_data_item = DataItem.DataItem(data.copy())
            document_model.append_data_item(computed_data_item)
            document_model.set_data_item_computation(computed_data_item, computation)
            document_model.recompute_all()
            self.assertTrue(numpy.array_equal(data_item.data, data))
            self.assertTrue(numpy.array_equal(computed_data_item.data, -data_item.data))
            data_item.set_data(((numpy.random.randn(10, 8) + 1) * 10).astype(numpy.uint32))
            document_model.recompute_all()
            self.assertFalse(numpy.array_equal(data_item.data, data))
            self.assertTrue(numpy.array_equal(computed_data_item.data, -data_item.data))

    def test_computation_is_live_after_copying_data_item_with_computation(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            data = ((numpy.random.randn(10, 8) + 1) * 10).astype(numpy.int32)
            data_item = DataItem.DataItem(data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("-a.xdata"))
            computation.create_object("a", document_model.get_object_specifier(data_item))
            computed_data_item = DataItem.DataItem(data.copy())
            document_model.append_data_item(computed_data_item)
            document_model.set_data_item_computation(computed_data_item, computation)
            document_model.recompute_all()
            self.assertTrue(numpy.array_equal(computed_data_item.data, -data_item.data))
            copied_data_item = document_model.copy_data_item(computed_data_item)
            data_item.set_data(((numpy.random.randn(10, 8) + 1) * 10).astype(numpy.int32))
            document_model.recompute_all()
            self.assertTrue(numpy.array_equal(computed_data_item.data, -data_item.data))
            self.assertTrue(numpy.array_equal(copied_data_item.data, -data_item.data))

    def test_computation_extracts_data_property_of_data_item(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            data = ((numpy.random.randn(10, 8) + 1) * 10).astype(numpy.uint32)
            data_item = DataItem.DataItem(data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("a.xdata"))
            computation.create_object("a", document_model.get_object_specifier(data_item))
            computed_data_item = DataItem.DataItem(data.copy())
            document_model.append_data_item(computed_data_item)
            document_model.set_data_item_computation(computed_data_item, computation)
            document_model.recompute_all()
            self.assertTrue(numpy.array_equal(computed_data_item.data, data_item.data))

    def test_resample_produces_correct_data(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            data = ((numpy.abs(numpy.random.randn(10, 8)) + 1) * 10).astype(numpy.uint32)
            data_item = DataItem.DataItem(data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("xd.resample_image(a.xdata, (5, 4))"))
            computation.create_object("a", document_model.get_object_specifier(data_item))
            computed_data_item = DataItem.DataItem(data.copy())
            document_model.append_data_item(computed_data_item)
            document_model.set_data_item_computation(computed_data_item, computation)
            document_model.recompute_all()
            self.assertTrue(numpy.array_equal(computed_data_item.data, Image.scaled(data_item.data, (5, 4))))

    def test_resample_with_data_shape_produces_correct_data(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            data = ((numpy.abs(numpy.random.randn(10, 8)) + 1) * 10).astype(numpy.uint32)
            data_item = DataItem.DataItem(data)
            document_model.append_data_item(data_item)
            data2 = numpy.zeros((5, 4), numpy.uint32)
            data_item2 = DataItem.DataItem(data2)
            document_model.append_data_item(data_item2)
            computation = document_model.create_computation(Symbolic.xdata_expression("xd.resample_image(a.xdata, b.xdata.data_shape)"))
            computation.create_object("a", document_model.get_object_specifier(data_item))
            computation.create_object("b", document_model.get_object_specifier(data_item2))
            computed_data_item = DataItem.DataItem(data.copy())
            document_model.append_data_item(computed_data_item)
            document_model.set_data_item_computation(computed_data_item, computation)
            document_model.recompute_all()
            self.assertTrue(numpy.array_equal(computed_data_item.data, Image.scaled(data_item.data, (5, 4))))

    def test_computation_extracts_display_data_property_of_data_item(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            src_data = ((numpy.random.randn(10, 8) + 1) * 10).astype(numpy.uint32)
            data_item = DataItem.DataItem(src_data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("a.display_xdata"))
            computation.create_object("a", document_model.get_object_specifier(data_item))
            data_node_dict = computation.write_to_dict()
            computation2 = document_model.create_computation()
            computation2.read_from_dict(data_node_dict)
            computation2.needs_update = True
            computation2.bind(document_model)
            data = DocumentModel.evaluate_data(computation).data
            data2 = DocumentModel.evaluate_data(computation2).data
            assert numpy.array_equal(data, src_data)
            assert numpy.array_equal(data, data2)

    def test_evaluation_with_variable_produces_correct_data(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            d = numpy.random.randn(2, 2)
            data_item = DataItem.DataItem(d)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("a.xdata + x"))
            computation.create_variable("x", value_type="integral", value=5)
            computation.create_object("a", document_model.get_object_specifier(data_item))
            data = DocumentModel.evaluate_data(computation).data
            assert numpy.array_equal(data, d + 5)

    def test_evaluation_with_two_variables_produces_correct_data(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            src_data = ((numpy.abs(numpy.random.randn(10, 8)) + 1) * 10).astype(numpy.uint32)
            data_item = DataItem.DataItem(src_data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("xd.gaussian_blur(a.xdata, x - y)"))
            computation.create_variable("x", value_type="integral", value=5)
            computation.create_variable("y", value_type="integral", value=5)
            computation.create_object("a", document_model.get_object_specifier(data_item))
            assert numpy.array_equal(DocumentModel.evaluate_data(computation).data, src_data)

    def test_changing_variable_value_updates_computation(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            src_data = numpy.random.randn(2, 2)
            data_item = DataItem.DataItem(src_data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("a.xdata + x"))
            x = computation.create_variable("x", value_type="integral", value=5)
            computation.create_object("a", document_model.get_object_specifier(data_item))
            computed_data_item = DataItem.DataItem(src_data.copy())
            document_model.append_data_item(computed_data_item)
            document_model.set_data_item_computation(computed_data_item, computation)
            document_model.recompute_all()
            self.assertTrue(numpy.array_equal(computed_data_item.data, src_data + 5))
            x.value = 8
            document_model.recompute_all()
            self.assertTrue(numpy.array_equal(computed_data_item.data, src_data + 8))

    def test_changing_region_property_updates_computation(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            src_data = numpy.random.randn(20, 20)
            data_item = DataItem.DataItem(src_data)
            line_region = Graphics.LineProfileGraphic()
            line_region.start = 0.25, 0.25
            line_region.end = 0.75, 0.75
            data_item.displays[0].add_graphic(line_region)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("xd.line_profile(src.display_xdata, line_region.vector, line_region.line_width)"))
            computation.create_object("src", document_model.get_object_specifier(data_item))
            computation.create_object("line_region", document_model.get_object_specifier(line_region))
            computed_data_item = DataItem.DataItem(src_data.copy())
            document_model.append_data_item(computed_data_item)
            document_model.set_data_item_computation(computed_data_item, computation)
            document_model.recompute_all()
            self.assertTrue(numpy.array_equal(computed_data_item.data, Core.function_line_profile(data_item.xdata, line_region.vector, 1.0).data))
            line_region.start = 0.25, 0.20
            document_model.recompute_all()
            self.assertTrue(numpy.array_equal(computed_data_item.data, Core.function_line_profile(data_item.xdata, line_region.vector, 1.0).data))

    def test_changing_variable_name_has_no_effect_on_computation(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            src_data = numpy.random.randn(2, 2)
            data_item = DataItem.DataItem(src_data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("a.xdata + x"))
            x = computation.create_variable("x", value_type="integral", value=5)
            computation.create_object("a", document_model.get_object_specifier(data_item))
            computed_data_item = DataItem.DataItem(src_data.copy())
            document_model.append_data_item(computed_data_item)
            document_model.set_data_item_computation(computed_data_item, computation)
            document_model.recompute_all()
            self.assertTrue(numpy.array_equal(computed_data_item.data, src_data + 5))
            x.name = "xx"
            document_model.recompute_all()
            self.assertTrue(numpy.array_equal(computed_data_item.data, src_data + 5))

    def test_computation_with_variable_reloads(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            src_data = ((numpy.abs(numpy.random.randn(10, 8)) + 1) * 10).astype(numpy.uint32)
            data_item = DataItem.DataItem(src_data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("a.xdata + x"))
            computation.create_variable("x", value_type="integral", value=5)
            computation.create_object("a", document_model.get_object_specifier(data_item))
            d = computation.write_to_dict()
            computation2 = document_model.create_computation()
            computation2.read_from_dict(d)
            computation2.needs_update = True
            computation2.bind(document_model)
            self.assertTrue(numpy.array_equal(DocumentModel.evaluate_data(computation).data, src_data + 5))
            self.assertTrue(numpy.array_equal(DocumentModel.evaluate_data(computation2).data, src_data + 5))

    def test_computation_variable_writes_and_reads(self):
        variable = Symbolic.ComputationVariable("x", value_type="integral", value=5)
        self.assertEqual(variable.name, "x")
        self.assertEqual(variable.value, 5)
        data_node_dict = variable.write_to_dict()
        variable2 = Symbolic.ComputationVariable()
        variable2.read_from_dict(data_node_dict)
        self.assertEqual(variable.name, variable2.name)
        self.assertEqual(variable.value, variable2.value)

    def test_computation_variable_change_type(self):
        variable = Symbolic.ComputationVariable("x", value_type="integral", value=5)
        variable.variable_type = "data_item"
        variable = Symbolic.ComputationVariable("x", value_type="integral", value=5)
        variable.variable_type = "region"

    def test_computation_reparsing_keeps_variables(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            src_data = ((numpy.random.randn(10, 8) + 1) * 10).astype(numpy.uint32)
            data_item = DataItem.DataItem(src_data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("a.xdata + x"))
            x = computation.create_variable("x", value_type="integral", value=5)
            computation.create_object("a", document_model.get_object_specifier(data_item))
            self.assertTrue(numpy.array_equal(DocumentModel.evaluate_data(computation).data, src_data + 5))
            computation.expression = Symbolic.xdata_expression("x + a.xdata")
            self.assertTrue(numpy.array_equal(DocumentModel.evaluate_data(computation).data, src_data + 5))

    def test_computation_using_object_parses_and_evaluates(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            src_data = ((numpy.random.randn(10, 8) + 1) * 10).astype(numpy.uint32)
            data_item = DataItem.DataItem(src_data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("a.xdata + x"))
            computation.create_variable("x", value_type="integral", value=5)
            computation.create_object("a", document_model.get_object_specifier(data_item))
            self.assertTrue(numpy.array_equal(DocumentModel.evaluate_data(computation).data, src_data + 5))

    def test_computation_using_object_updates_when_data_changes(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            src_data = ((numpy.random.randn(10, 8) + 1) * 10).astype(numpy.uint32)
            data_item = DataItem.DataItem(src_data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("a.xdata + x"))
            computation.create_variable("x", value_type="integral", value=5)
            computation.create_object("a", document_model.get_object_specifier(data_item))
            self.assertTrue(numpy.array_equal(DocumentModel.evaluate_data(computation).data, src_data + 5))
            d = computation.write_to_dict()
            read_computation = document_model.create_computation()
            read_computation.read_from_dict(d)
            read_computation.needs_update = True
            read_computation.bind(document_model)
            src_data2 = ((numpy.random.randn(10, 8) + 1) * 10).astype(numpy.uint32)
            data_item.set_data(src_data2)
            self.assertTrue(numpy.array_equal(DocumentModel.evaluate_data(read_computation).data, src_data2 + 5))

    def test_computation_using_object_updates_efficiently_when_region_changes(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            src_data = ((numpy.abs(numpy.random.randn(12, 8)) + 1) * 10).astype(numpy.uint32)
            data_item = DataItem.DataItem(src_data)
            region = Graphics.RectangleGraphic()
            region.bounds = Geometry.FloatRect.from_center_and_size(Geometry.FloatPoint(0.5, 0.5), Geometry.FloatSize(0.5, 0.5))
            data_item.displays[0].add_graphic(region)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("xd.crop(a.xdata, r.bounds)"))
            computation.create_object("a", document_model.get_object_specifier(data_item))
            computation.create_object("r", document_model.get_object_specifier(region))
            computed_data_item = DataItem.DataItem(src_data.copy())
            document_model.append_data_item(computed_data_item)
            document_model.set_data_item_computation(computed_data_item, computation)
            document_model.recompute_all()
            self.assertTrue(numpy.array_equal(computed_data_item.data, src_data[3:9, 2:6]))
            region.bounds = Geometry.FloatRect.from_center_and_size(Geometry.FloatPoint(0.25, 0.25), Geometry.FloatSize(0.5, 0.5))
            region.bounds = Geometry.FloatRect.from_center_and_size(Geometry.FloatPoint(0.0, 0.0), Geometry.FloatSize(0.5, 0.5))
            region.bounds = Geometry.FloatRect.from_center_and_size(Geometry.FloatPoint(0.25, 0.25), Geometry.FloatSize(0.5, 0.5))
            evaluation_count = computation._evaluation_count_for_test
            document_model.recompute_all()
            self.assertEqual(computation._evaluation_count_for_test - evaluation_count, 1)
            self.assertTrue(numpy.array_equal(computed_data_item.data, src_data[0:6, 0:4]))

    def test_computation_updates_efficiently_when_variable_changes(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            src_data = ((numpy.abs(numpy.random.randn(12, 8)) + 1) * 10).astype(numpy.uint32)
            data_item = DataItem.DataItem(src_data)
            region = Graphics.RectangleGraphic()
            region.bounds = Geometry.FloatRect.from_center_and_size(Geometry.FloatPoint(0.5, 0.5), Geometry.FloatSize(0.5, 0.5))
            data_item.displays[0].add_graphic(region)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("xd.gaussian_blur(a.xdata, s)"))
            s = computation.create_variable("s", value_type="integral", value=5)
            computation.create_object("a", document_model.get_object_specifier(data_item))
            computed_data_item = DataItem.DataItem(src_data.copy())
            document_model.append_data_item(computed_data_item)
            document_model.set_data_item_computation(computed_data_item, computation)
            document_model.recompute_all()
            evaluation_count = computation._evaluation_count_for_test
            s.value = 4
            self.assertEqual(computation._evaluation_count_for_test - evaluation_count, 0)
            document_model.recompute_all()
            self.assertEqual(computation._evaluation_count_for_test - evaluation_count, 1)

    def test_computation_updates_efficiently_when_variable_added_or_removed(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            src_data = ((numpy.abs(numpy.random.randn(12, 8)) + 1) * 10).astype(numpy.uint32)
            data_item = DataItem.DataItem(src_data)
            region = Graphics.RectangleGraphic()
            region.bounds = Geometry.FloatRect.from_center_and_size(Geometry.FloatPoint(0.5, 0.5), Geometry.FloatSize(0.5, 0.5))
            data_item.displays[0].add_graphic(region)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("xd.gaussian_blur(a.xdata, s)"))
            s = computation.create_variable("s", value_type="integral", value=5)
            computation.create_object("a", document_model.get_object_specifier(data_item))
            computed_data_item = DataItem.DataItem(src_data.copy())
            document_model.append_data_item(computed_data_item)
            document_model.set_data_item_computation(computed_data_item, computation)
            document_model.recompute_all()
            evaluation_count = computation._evaluation_count_for_test
            t = computation.create_variable("t", value_type="integral", value=5)
            self.assertEqual(computation._evaluation_count_for_test - evaluation_count, 0)
            document_model.recompute_all()
            self.assertEqual(computation._evaluation_count_for_test - evaluation_count, 1)
            computation.remove_variable(t)
            self.assertEqual(computation._evaluation_count_for_test - evaluation_count, 1)
            document_model.recompute_all()
            self.assertEqual(computation._evaluation_count_for_test - evaluation_count, 2)

    def test_computation_updates_efficiently_when_expression_changes(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            src_data = ((numpy.abs(numpy.random.randn(12, 8)) + 1) * 10).astype(numpy.uint32)
            data_item = DataItem.DataItem(src_data)
            region = Graphics.RectangleGraphic()
            region.bounds = Geometry.FloatRect.from_center_and_size(Geometry.FloatPoint(0.5, 0.5), Geometry.FloatSize(0.5, 0.5))
            data_item.displays[0].add_graphic(region)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("xd.gaussian_blur(a.xdata, s)"))
            s = computation.create_variable("s", value_type="integral", value=5)
            computation.create_object("a", document_model.get_object_specifier(data_item))
            computed_data_item = DataItem.DataItem(src_data.copy())
            document_model.append_data_item(computed_data_item)
            document_model.set_data_item_computation(computed_data_item, computation)
            document_model.recompute_all()
            evaluation_count = computation._evaluation_count_for_test
            computation.expression = Symbolic.xdata_expression("xd.gaussian_blur(a, 2)")
            # computation should not be re-evaluated until requested
            self.assertEqual(computation._evaluation_count_for_test - evaluation_count, 0)
            document_model.recompute_all()
            self.assertEqual(computation._evaluation_count_for_test - evaluation_count, 1)

    def test_computation_with_object_writes_and_reads(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            src_data = ((numpy.random.randn(10, 8) + 1) * 10).astype(numpy.uint32)
            data_item = DataItem.DataItem(src_data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("a.xdata + 4"))
            a_specifier = document_model.get_object_specifier(data_item)
            computation.create_object("a", document_model.get_object_specifier(data_item))
            d = computation.write_to_dict()
            read_computation = document_model.create_computation()
            read_computation.read_from_dict(d)
            self.assertEqual(read_computation.variables[0].name, "a")
            self.assertEqual(read_computation.variables[0].specifier, a_specifier)

    def test_computation_with_object_reloads(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            src_data = ((numpy.random.randn(10, 8) + 1) * 10).astype(numpy.uint32)
            data_item = DataItem.DataItem(src_data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("a.xdata + x"))
            computation.create_variable("x", value_type="integral", value=5)
            computation.create_object("a", document_model.get_object_specifier(data_item))
            d = computation.write_to_dict()
            computation2 = document_model.create_computation()
            computation2.read_from_dict(d)
            computation2.needs_update = True
            computation2.bind(document_model)
            self.assertTrue(numpy.array_equal(DocumentModel.evaluate_data(computation).data, src_data + 5))
            self.assertTrue(numpy.array_equal(DocumentModel.evaluate_data(computation2).data, src_data + 5))

    def test_computation_with_object_evaluates_correctly_after_changing_the_variable_name(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            src_data = ((numpy.random.randn(10, 8) + 1) * 10).astype(numpy.uint32)
            data_item = DataItem.DataItem(src_data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("a.xdata + x"))
            computation.create_object("a", document_model.get_object_specifier(data_item))
            x = computation.create_variable("x", value_type="integral", value=5)
            self.assertTrue(numpy.array_equal(DocumentModel.evaluate_data(computation).data, src_data + 5))
            x.name = "xx"
            computation.expression = Symbolic.xdata_expression("a.xdata + xx")
            self.assertTrue(numpy.array_equal(DocumentModel.evaluate_data(computation).data, src_data + 5))

    def test_computation_with_object_evaluates_correctly_after_changing_the_specifier(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            src_data1 = ((numpy.random.randn(10, 8) + 1) * 10).astype(numpy.uint32)
            src_data2 = ((numpy.random.randn(10, 8) + 1) * 10).astype(numpy.uint32)
            data_item1 = DataItem.DataItem(src_data1)
            data_item2 = DataItem.DataItem(src_data2)
            document_model.append_data_item(data_item1)
            document_model.append_data_item(data_item2)
            expression = "a.xdata + 1"
            computation = document_model.create_computation(Symbolic.xdata_expression(expression))
            a = computation.create_object("a", document_model.get_object_specifier(data_item1))
            self.assertTrue(numpy.array_equal(DocumentModel.evaluate_data(computation).data, src_data1 + 1))
            a.specifier = document_model.get_object_specifier(data_item2)
            computation.expression = Symbolic.xdata_expression(expression)
            self.assertTrue(numpy.array_equal(DocumentModel.evaluate_data(computation).data, src_data2 + 1))

    def test_computation_fires_needs_update_event_when_specifier_changes(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            src_data1 = ((numpy.random.randn(10, 8) + 1) * 10).astype(numpy.uint32)
            src_data2 = ((numpy.random.randn(10, 8) + 1) * 10).astype(numpy.uint32)
            data_item1 = DataItem.DataItem(src_data1)
            data_item2 = DataItem.DataItem(src_data2)
            document_model.append_data_item(data_item1)
            document_model.append_data_item(data_item2)
            expression = "a.xdata + 1"
            computation = document_model.create_computation(Symbolic.xdata_expression(expression))
            a = computation.create_object("a", document_model.get_object_specifier(data_item1))
            needs_update_ref = [False]
            def needs_update():
                needs_update_ref[0] = True
            needs_update_event_listener = computation.computation_mutated_event.listen(needs_update)
            with contextlib.closing(needs_update_event_listener):
                a.specifier = document_model.get_object_specifier(data_item2)
            self.assertTrue(needs_update_ref[0])

    def test_computation_in_document_recomputes_when_specifier_changes(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            src_data1 = ((numpy.random.randn(10, 8) + 1) * 10).astype(numpy.uint32)
            src_data2 = ((numpy.random.randn(10, 8) + 1) * 10).astype(numpy.uint32)
            data_item1 = DataItem.DataItem(src_data1)
            data_item2 = DataItem.DataItem(src_data2)
            document_model.append_data_item(data_item1)
            document_model.append_data_item(data_item2)
            expression = "a.xdata + 1"
            computation = document_model.create_computation(Symbolic.xdata_expression(expression))
            a = computation.create_object("a", document_model.get_object_specifier(data_item1))
            computed_data_item = DataItem.DataItem(src_data1.copy())
            document_model.append_data_item(computed_data_item)
            document_model.set_data_item_computation(computed_data_item, computation)
            document_model.recompute_all()
            self.assertTrue(numpy.array_equal(computed_data_item.data, src_data1 + 1))
            a.specifier = document_model.get_object_specifier(data_item2)
            document_model.recompute_all()
            self.assertTrue(numpy.array_equal(computed_data_item.data, src_data2 + 1))

    def test_computation_in_document_is_still_live_when_region_specifier_uuid_str_changes(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            src_data = ((numpy.abs(numpy.random.randn(10, 10)) + 1) * 10).astype(numpy.uint32)
            data_item = DataItem.DataItem(src_data)
            region1 = Graphics.RectangleGraphic()
            region1.bounds = Geometry.FloatRect.from_tlhw(0.0, 0.0, 0.5, 0.5)
            data_item.displays[0].add_graphic(region1)
            region2 = Graphics.RectangleGraphic()
            region2.bounds = Geometry.FloatRect.from_tlhw(0.5, 0.5, 0.5, 0.5)
            data_item.displays[0].add_graphic(region2)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("xd.crop(a.xdata, r.bounds)"))
            computation.create_object("a", document_model.get_object_specifier(data_item))
            r = computation.create_object("r", document_model.get_object_specifier(region1))
            computed_data_item = DataItem.DataItem(src_data.copy())
            document_model.append_data_item(computed_data_item)
            document_model.set_data_item_computation(computed_data_item, computation)
            # verify assumptions
            document_model.recompute_all()
            self.assertTrue(numpy.array_equal(computed_data_item.data, src_data[0:5, 0:5]))
            # now switch the region uuid
            r.specifier_uuid_str = str(region2.uuid)
            document_model.recompute_all()
            self.assertTrue(numpy.array_equal(computed_data_item.data, src_data[5:10, 5:10]))
            # and make sure recompute happens when new region uuid changes
            region2.bounds = Geometry.FloatRect.from_tlhw(0.0, 0.0, 0.5, 0.5)
            document_model.recompute_all()
            self.assertTrue(numpy.array_equal(computed_data_item.data, src_data[0:5, 0:5]))

    def test_computation_with_raw_reference_copies(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            src_data = ((numpy.random.randn(10, 8) + 1) * 10).astype(numpy.uint32)
            data_item = DataItem.DataItem(src_data)
            region = Graphics.RectangleGraphic()
            data_item.displays[0].add_graphic(region)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("a.xdata + x"))
            computation.create_variable("x", value_type="integral", value=5)
            a = computation.create_object("a", document_model.get_object_specifier(data_item))
            computation.remove_variable(a)
            copy.deepcopy(computation)

    def test_evaluation_error_recovers_gracefully(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            src_data = ((numpy.abs(numpy.random.randn(12, 8)) + 1) * 10).astype(numpy.uint32)
            data_item = DataItem.DataItem(src_data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("xd.line_profile(a.xdata, xd.vector(xd.norm_point(0.25, 0.25), xd.norm_point(0.5, 0.5)), x)"))
            x = computation.create_variable("x", value_type="integral", value=0)
            computation.create_object("a", document_model.get_object_specifier(data_item))
            computed_data_item = DataItem.DataItem(src_data.copy())
            document_model.append_data_item(computed_data_item)
            document_model.set_data_item_computation(computed_data_item, computation)
            document_model.recompute_all()
            self.assertIsNotNone(computation.error_text)
            self.assertEqual(len(computed_data_item.data.shape), 2)  # original data
            x.value = 1
            document_model.recompute_all()
            self.assertIsNone(computation.error_text)
            self.assertIsNotNone(computed_data_item.data)
            self.assertEqual(len(computed_data_item.data.shape), 1)  # computed data

    def test_various_expressions_produces_data(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            src_data = ((numpy.abs(numpy.random.randn(10, 8)) + 1) * 10).astype(numpy.uint32)
            data_item = DataItem.DataItem(src_data)
            document_model.append_data_item(data_item)
            script_and_data = [
                ("xd.histogram(a.xdata, 10)", None),
                ("xd.line_profile(a.xdata, xd.vector(xd.norm_point(0.1, 0.1), xd.norma_point(0.8, 0.7)), 10)", None),
                ("xd.transpose_flip(a.xdata, False, True, False)", None),
                ("xd.crop(a.xdata, xd.rectangle_from_origin_size(xd.norm_point(0, 0), xd.norm_size(0.5, 0.625)))", src_data[0:5, 0:5]),
                ("xd.sum(a.xdata, 0)", numpy.sum(src_data, 0)),
                ("xd.resample_image(a.xdata, (32, 32))", Image.scaled(src_data, (32, 32))),
                ("xd.resample_image(a.xdata, a.xdata.data_shape)", src_data),
                ("xd.resample_image(a.xdata, xd.crop(a.xdata, xd.rectangle_from_origin_size(xd.norm_point(0, 0), xd.norm_size(0.5, 0.625))).data_shape)", Image.scaled(src_data, (5, 5))),
                ("a.xdata + x", src_data + 5),
                ("xd.gaussian_blur(a.xdata, x + x)", None),
                ("xd.gaussian_blur(a.xdata, x - 2)", None),
                ("xd.gaussian_blur(a.xdata, 2 * x)", None),
                ("xd.gaussian_blur(a.xdata, +x)", None),
            ]
            for script, data in script_and_data:
                computation = document_model.create_computation(Symbolic.xdata_expression(script))
                computation.create_object("a", document_model.get_object_specifier(data_item))
                computation.create_variable("x", value_type="integral", value=5)
                computed_data_item = DataItem.DataItem(src_data.copy())
                document_model.append_data_item(computed_data_item)
                document_model.set_data_item_computation(computed_data_item, computation)
                document_model.recompute_all()
                self.assertIsNotNone(computed_data_item.data)
                if data is not None:
                    self.assertTrue(numpy.array_equal(computed_data_item.data, data))

    def test_conversion_to_int(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            src_data = ((numpy.abs(numpy.random.randn(10, 8)) + 1) * 10).astype(numpy.float64)
            data_item = DataItem.DataItem(src_data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation()
            computation.create_object("src", document_model.get_object_specifier(data_item))
            computation.expression = Symbolic.xdata_expression("xd.astype(src.xdata, int)")
            data = DocumentModel.evaluate_data(computation).data
            self.assertEqual(data.dtype, numpy.int_)
            self.assertTrue(numpy.array_equal(data, src_data.astype(int)))
            computation.expression = Symbolic.xdata_expression("xd.astype(src.xdata, numpy.int16)")
            data = DocumentModel.evaluate_data(computation).data
            self.assertEqual(data.dtype, numpy.int16)
            self.assertTrue(numpy.array_equal(data, src_data.astype(numpy.int16)))
            computation.expression = Symbolic.xdata_expression("xd.astype(src.xdata, numpy.int32)")
            data = DocumentModel.evaluate_data(computation).data
            self.assertEqual(data.dtype, numpy.int32)
            self.assertTrue(numpy.array_equal(data, src_data.astype(numpy.int32)))
            computation.expression = Symbolic.xdata_expression("xd.astype(src.xdata, numpy.int64)")
            data = DocumentModel.evaluate_data(computation).data
            self.assertEqual(data.dtype, numpy.int64)
            self.assertTrue(numpy.array_equal(data, src_data.astype(numpy.int64)))

    def test_conversion_to_uint(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            src_data = ((numpy.abs(numpy.random.randn(10, 8)) + 1) * 10).astype(numpy.float64)
            data_item = DataItem.DataItem(src_data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation()
            computation.create_object("src", document_model.get_object_specifier(data_item))
            computation.expression = Symbolic.xdata_expression("xd.astype(src.xdata, numpy.uint8)")
            data = DocumentModel.evaluate_data(computation).data
            self.assertEqual(data.dtype, numpy.uint8)
            self.assertTrue(numpy.array_equal(data, src_data.astype(numpy.uint8)))
            computation.expression = Symbolic.xdata_expression("xd.astype(src.xdata, numpy.uint16)")
            data = DocumentModel.evaluate_data(computation).data
            self.assertEqual(data.dtype, numpy.uint16)
            self.assertTrue(numpy.array_equal(data, src_data.astype(numpy.uint16)))
            computation.expression = Symbolic.xdata_expression("xd.astype(src.xdata, numpy.uint32)")
            data = DocumentModel.evaluate_data(computation).data
            self.assertEqual(data.dtype, numpy.uint32)
            self.assertTrue(numpy.array_equal(data, src_data.astype(numpy.uint32)))
            computation.expression = Symbolic.xdata_expression("xd.astype(src.xdata, numpy.uint64)")
            data = DocumentModel.evaluate_data(computation).data
            self.assertEqual(data.dtype, numpy.uint64)
            self.assertTrue(numpy.array_equal(data, src_data.astype(numpy.uint64)))

    def test_conversion_to_float(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            src_data = ((numpy.abs(numpy.random.randn(10, 8)) + 1) * 10).astype(numpy.int32)
            data_item = DataItem.DataItem(src_data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation()
            computation.create_object("src", document_model.get_object_specifier(data_item))
            computation.expression = Symbolic.xdata_expression("xd.astype(src.xdata, numpy.float32)")
            data = DocumentModel.evaluate_data(computation).data
            self.assertEqual(data.dtype, numpy.float32)
            self.assertTrue(numpy.array_equal(data, src_data.astype(numpy.float32)))
            computation.expression = Symbolic.xdata_expression("xd.astype(src.xdata, numpy.float64)")
            data = DocumentModel.evaluate_data(computation).data
            self.assertEqual(data.dtype, numpy.float64)
            self.assertTrue(numpy.array_equal(data, src_data.astype(numpy.float64)))

    def test_conversion_to_complex(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            src_data = ((numpy.abs(numpy.random.randn(10, 8)) + 1) * 10).astype(numpy.int32)
            data_item = DataItem.DataItem(src_data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation()
            computation.create_object("src", document_model.get_object_specifier(data_item))
            computation.expression = Symbolic.xdata_expression("xd.astype(src.xdata, numpy.complex64)")
            data = DocumentModel.evaluate_data(computation).data
            self.assertEqual(data.dtype, numpy.complex64)
            self.assertTrue(numpy.array_equal(data, src_data.astype(numpy.complex64)))
            computation.expression = Symbolic.xdata_expression("xd.astype(src.xdata, numpy.complex128)")
            data = DocumentModel.evaluate_data(computation).data
            self.assertEqual(data.dtype, numpy.complex128)
            self.assertTrue(numpy.array_equal(data, src_data.astype(numpy.complex128)))

    def test_data_descriptor_is_maintained_during_evaluate(self):
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            src_data1 = ((numpy.abs(numpy.random.randn(10,)) + 1) * 10).astype(numpy.int32)
            data_item1 = DataItem.DataItem(src_data1)
            document_model.append_data_item(data_item1)
            src_data2 = ((numpy.abs(numpy.random.randn(10,)) + 1) * 10).astype(numpy.int32)
            data_item2 = DataItem.DataItem(src_data2)
            document_model.append_data_item(data_item2)
            computation = document_model.create_computation()
            computation.create_object("src1", document_model.get_object_specifier(data_item1))
            computation.create_object("src2", document_model.get_object_specifier(data_item2))
            computation.expression = Symbolic.xdata_expression("xd.vstack((src1.xdata, src2.xdata))")
            data_and_metadata = DocumentModel.evaluate_data(computation)
            self.assertEqual(data_and_metadata.collection_dimension_count, 1)
            self.assertEqual(data_and_metadata.datum_dimension_count, 1)

    def test_computation_evaluates_on_thread(self):
        document_model = DocumentModel.DocumentModel()
        # in order to make this test go fast, attach the call_soon event here
        # and handle it, setting the continue event at the same time.
        # this must go before the document controller is created since call_soon
        # is a 'fire_any' style event, meaning that only the first handler takes
        # it. we return False to let the document controller eventually handle it.
        continue_event = threading.Event()
        def do_call_soon(fn):
            continue_event.set()
            return False
        listener = document_model.call_soon_event.listen(do_call_soon)
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            src_data = numpy.random.randn(2, 2)
            data_item = DataItem.DataItem(src_data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("-a.xdata"))
            computation.create_object("a", document_model.get_object_specifier(data_item))
            computed_data_item = DataItem.DataItem(src_data.copy())
            document_model.append_data_item(computed_data_item)
            document_model.set_data_item_computation(computed_data_item, computation)
            document_model.start_dispatcher()
            continue_event.wait(10.0)
            listener.close()
            listener = None
            document_controller.periodic()
            self.assertTrue(numpy.array_equal(computed_data_item.data, -src_data))

    def test_computation_on_deleted_data_item(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            src_data = numpy.random.randn(2, 2)
            data_item = DataItem.DataItem(src_data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("-a.xdata"))
            computation.create_object("a", document_model.get_object_specifier(data_item))
            computed_data_item = DataItem.DataItem(src_data.copy())
            document_model.append_data_item(computed_data_item)
            document_model.set_data_item_computation(computed_data_item, computation)
            document_model.remove_data_item(data_item)
            document_model.recompute_all()

    def test_computation_updates_timezone(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            try:
                src_data = numpy.random.randn(2, 2)
                data_item = DataItem.DataItem(src_data)
                document_model.append_data_item(data_item)
                computation = document_model.create_computation(Symbolic.xdata_expression("-a.xdata"))
                computation.create_object("a", document_model.get_object_specifier(data_item))
                Utility.local_timezone_override = [None]
                Utility.local_utcoffset_override = [0]
                computed_data_item = DataItem.DataItem(src_data.copy())
                document_model.append_data_item(computed_data_item)
                document_model.set_data_item_computation(computed_data_item, computation)
                self.assertIsNone(computed_data_item.timezone)
                self.assertEqual(computed_data_item.timezone_offset, "+0000")
                Utility.local_timezone_override = ["Europe/Athens"]
                Utility.local_utcoffset_override = [180]
                document_model.recompute_all()
                self.assertEqual(computed_data_item.timezone, "Europe/Athens")
                self.assertEqual(computed_data_item.timezone_offset, "+0300")
            finally:
                Utility.local_timezone_override = None
                Utility.local_utcoffset_override = None

    def test_deleting_one_variable_on_bound_computation_rebinds_properly(self):
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            src_data = numpy.random.randn(2, 2)
            data_item = DataItem.DataItem(src_data)
            document_model.append_data_item(data_item)
            computation = document_model.create_computation(Symbolic.xdata_expression("x"))
            x_var = computation.create_variable("x")
            computed_data_item = DataItem.DataItem(src_data.copy())
            document_model.append_data_item(computed_data_item)
            document_model.set_data_item_computation(computed_data_item, computation)
            computation.remove_variable(x_var)

    class ComputeExecError:
        def __init__(self, computation, **kwargs):
            self.computation = computation

        def execute(self, src_xdata):
            raise RuntimeError()

        def commit(self):
            pass

    def test_computation_error_is_handled_in_main_thread(self):
        Symbolic.register_computation_type("compute_error", self.ComputeExecError)
        document_model = DocumentModel.DocumentModel()
        # in order to make this test go fast, attach the call_soon event here
        # and handle it, setting the continue event at the same time.
        # this must go before the document controller is created since call_soon
        # is a 'fire_any' style event, meaning that only the first handler takes
        # it. we return False to let the document controller eventually handle it.
        continue_event = threading.Event()
        def do_call_soon(fn):
            continue_event.set()
            return False
        listener = document_model.call_soon_event.listen(do_call_soon)
        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            data_item = DataItem.DataItem(numpy.zeros((2, 2)))
            document_model.append_data_item(data_item)
            computation1 = document_model.create_computation()
            computation1.create_object("src_xdata", document_model.get_object_specifier(data_item, "xdata"))
            computation1.processing_id = "compute_error"
            document_model.append_computation(computation1)
            document_model.start_dispatcher()
            continue_event.wait(10.0)
            listener.close()
            document_controller.periodic()

    def test_computation_error_clear_initial_computation_flag(self):
        Symbolic.register_computation_type("compute_error", self.ComputeExecError)
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            data_item = DataItem.DataItem(numpy.zeros((2, 2)))
            document_model.append_data_item(data_item)
            computation1 = document_model.create_computation()
            computation1.create_object("src_xdata", document_model.get_object_specifier(data_item, "xdata"))
            computation1.processing_id = "compute_error"
            document_model.append_computation(computation1)
            document_model.recompute_all()
            self.assertTrue(computation1.is_initial_computation_complete.wait(0.01))

    def test_removing_computation_unbinds(self):
        Symbolic.register_computation_type("compute_error", self.ComputeExecError)
        document_model = DocumentModel.DocumentModel()
        with contextlib.closing(document_model):
            data_item = DataItem.DataItem(numpy.random.randn(2, 2))
            document_model.append_data_item(data_item)
            document_model.get_invert_new(data_item)
            computation = document_model.computations[0]
            self.assertIsNotNone(computation.variables[0].bound_item)
            document_model.remove_data_item(document_model.data_items[1])
            self.assertIsNone(computation.variables[0].bound_item)

    def disabled_test_reshape_rgb(self):
        assert False

    def disabled_test_computation_with_data_error_gets_reported(self):
        assert False  # when the data node returns None

    def disabled_test_computation_variable_gets_closed(self):
        assert False

    def disabled_test_computation_with_cycles_fails_gracefully(self):
        assert False

    def disabled_test_computations_handle_constant_values_as_errors(self):
        # computation.parse_expression(document_model, "7", dict())
        assert False

    def disabled_test_computations_update_data_item_dependencies_list(self):
        assert False

    def disabled_test_function_to_modify_intensity_calibration(self):
        assert False

    def disabled_test_function_to_modify_dimensional_calibrations(self):
        assert False

    def disabled_test_function_to_modify_metadata(self):
        assert False

    def disabled_test_data_slice_calibration_with_step(self):
        # d[::2, :, :]
        pass

    def disabled_test_invalid_computation_produces_error_message(self):
        # example d[3:3, ...]
        # should return None be allowed? what about raise exception in all cases?
        pass

if __name__ == '__main__':
    logging.getLogger().setLevel(logging.DEBUG)
    unittest.main()



# standard libraries
import contextlib
import unittest

# third party libraries
import numpy

# local libraries
from nion.swift import Application
from nion.swift import DocumentController
from nion.swift import Facade
from nion.swift import RecorderPanel
from nion.swift.model import DataItem
from nion.swift.model import DocumentModel
from nion.ui import TestUI


Facade.initialize()

class TestRecorderPanelClass(unittest.TestCase):

    def setUp(self):
        self.app = Application.Application(TestUI.UserInterface(), set_global=False)

    def tearDown(self):
        pass

    def test_recorder_records_live_data(self):
        app = Application.Application(TestUI.UserInterface(), set_global=False)
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            data_item = DataItem.DataItem(numpy.ones((8, 8)))
            document_model.append_data_item(data_item)
            recorder = RecorderPanel.Recorder(document_controller, data_item)
            with contextlib.closing(recorder):
                with document_model.data_item_live(data_item):
                    count = 4
                    recorder.start_recording(10, 1, count)
                    for i in range(count):
                        recorder.continue_recording(10 + i + 0.25)
                        recorder.continue_recording(10 + i + 0.75)
                        data_item.set_data(data_item.data + 1)
            self.assertEqual(2, len(document_model.data_items))
            recorded_data_item = document_model.data_items[1]
            self.assertTrue(recorded_data_item.xdata.is_sequence)
            self.assertEqual((4, 8, 8), recorded_data_item.xdata.dimensional_shape)

    def test_recorder_puts_recorded_data_item_under_transaction(self):
        app = Application.Application(TestUI.UserInterface(), set_global=False)
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            data_item = DataItem.DataItem(numpy.ones((8, 8)))
            document_model.append_data_item(data_item)
            recorder = RecorderPanel.Recorder(document_controller, data_item)
            with contextlib.closing(recorder):
                with document_model.data_item_live(data_item):
                    count = 4
                    recorder.start_recording(10, 1, count)
                    for i in range(count):
                        recorder.continue_recording(10 + i + 0.25)
                        data_item.set_data(data_item.data + 1)
                        if i != count - 1:  # last iteration it should fall out of transaction state since it is finished
                            self.assertTrue(document_model.data_items[1].in_transaction_state)
                        else:
                            self.assertFalse(document_model.data_items[1].in_transaction_state)
            self.assertFalse(document_model.data_items[1].in_transaction_state)

    def test_recorder_state_is_reported_properly(self):
        app = Application.Application(TestUI.UserInterface(), set_global=False)
        document_model = DocumentModel.DocumentModel()
        document_controller = DocumentController.DocumentController(app.ui, document_model, workspace_id="library")
        with contextlib.closing(document_controller):
            data_item = DataItem.DataItem(numpy.ones((8, 8)))
            document_model.append_data_item(data_item)
            recorder_state_ref = ["unknown"]
            def recorder_state_changed(recorder_state): recorder_state_ref[0] = recorder_state
            recorder = RecorderPanel.Recorder(document_controller, data_item)
            recorder.on_recording_state_changed = recorder_state_changed
            with contextlib.closing(recorder):
                with document_model.data_item_live(data_item):
                    count = 4
                    recorder.start_recording(10, 1, count)
                    for i in range(count):
                        recorder.continue_recording(10 + i + 0.25)
                        data_item.set_data(data_item.data + 1)
                        if i != count - 1:  # last iteration it should fall out of transaction state since it is finished
                            self.assertEqual(recorder_state_ref[0], "recording")
                        else:
                            self.assertEqual(recorder_state_ref[0], "stopped")
                self.assertEqual(recorder_state_ref[0], "stopped")



# standard libraries
import copy
import functools
import gettext
import json
import operator
import random
import string
import uuid
import weakref

# third party libraries
# None

# local libraries
from nion.swift import DataItemThumbnailWidget
from nion.swift import Undo
from nion.swift.model import DataItem
from nion.swift.model import Symbolic
from nion.ui import CanvasItem
from nion.ui import Dialog
from nion.utils import Binding
from nion.utils import Converter
from nion.utils import Event
from nion.utils import Geometry

_ = gettext.gettext


class ComputationModel:
    """Represents a computation. Tracks a display specifier for changes to it and its computation content.

    Provides read/write access to the computation_text via the property.

    Provides a computation_text_changed event, always fired on UI thread.
    """

    def __init__(self, document_controller):
        self.__weak_document_controller = weakref.ref(document_controller)
        self.__display_specifier = DataItem.DisplaySpecifier()
        self.__set_display_specifier(DataItem.DisplaySpecifier())
        self.__computation_changed_or_mutated_event_listener = None
        self.__computation_variable_inserted_event_listener = None
        self.__computation_variable_removed_event_listener = None
        self.__computation_property_changed_event_listener = None
        self.__computation_label = None
        self.__computation_text = None
        self.__error_text = None
        self.__object_property_changed_event_listeners = dict()
        self.__variable_property_changed_event_listeners = dict()
        self.computation_label_changed_event = Event.Event()
        self.computation_text_changed_event = Event.Event()
        self.error_text_changed_event = Event.Event()
        self.variable_inserted_event = Event.Event()
        self.variable_removed_event = Event.Event()
        self.variable_property_changed_event = Event.Event()

    def close(self):
        self.__set_display_specifier(DataItem.DisplaySpecifier())

    @property
    def document_controller(self):
        return self.__weak_document_controller()

    @property
    def display_specifier(self):
        return self.__display_specifier

    @property
    def __computation(self):
        if self.__display_specifier.data_item:
            return self.document_controller.document_model.get_data_item_computation(self.__display_specifier.data_item)
        return None

    @property
    def computation(self):
        return self.__computation

    def set_data_item(self, data_item):
        self.__set_display_specifier(DataItem.DisplaySpecifier.from_data_item(data_item))

    class AddVariableCommand(Undo.UndoableCommand):

        def __init__(self, document_model, computation: Symbolic.Computation, name: str=None, value_type: str=None, value=None, value_default=None, value_min=None, value_max=None, control_type: str=None, specifier: dict=None, label: str=None):
            super().__init__(_("Add Computation Variable"))
            self.__document_model = document_model
            self.__computation_uuid = computation.uuid
            self.__name = name
            self.__value_type = value_type
            self.__value = value
            self.__value_default = value_default
            self.__value_min = value_min
            self.__value_max = value_max
            self.__control_type = control_type
            self.__specifier = specifier
            self.__label = label
            self.__variable_index = None
            self.initialize()

        def close(self):
            self.__computation_uuid = None
            self.__name = None
            self.__value_type = None
            self.__value = None
            self.__value_default = None
            self.__value_min = None
            self.__value_max = None
            self.__control_type = None
            self.__specifier = None
            self.__label = None
            self.__variable = None
            super().close()

        def perform(self):
            computation = self.__document_model.get_computation_by_uuid(self.__computation_uuid)
            variable = computation.create_variable(self.__name, self.__value_type, self.__value, self.__value_default, self.__value_min, self.__value_max, self.__control_type, self.__specifier, self.__label)
            self.__variable_index = computation.variables.index(variable)

        def _get_modified_state(self):
            computation = self.__document_model.get_computation_by_uuid(self.__computation_uuid)
            return computation.modified_state

        def _set_modified_state(self, modified_state) -> None:
            computation = self.__document_model.get_computation_by_uuid(self.__computation_uuid)
            computation.modified_state = modified_state

        def _undo(self):
            computation = self.__document_model.get_computation_by_uuid(self.__computation_uuid)
            variable = computation.variables[self.__variable_index]
            computation.remove_variable(variable)

        def _redo(self):
            self.perform()

    class RemoveVariableCommand(Undo.UndoableCommand):

        def __init__(self, document_model, computation: Symbolic.Computation, variable: Symbolic.ComputationVariable):
            super().__init__(_("Remove Computation Variable"))
            self.__document_model = document_model
            self.__computation_uuid = computation.uuid
            self.__variable_index = computation.variables.index(variable)
            self.__variable_dict = variable.write_to_dict()
            self.initialize()

        def close(self):
            self.__computation_uuid = None
            self.__variable_index = None
            self.__variable_dict = None
            super().close()

        def perform(self):
            computation = self.__document_model.get_computation_by_uuid(self.__computation_uuid)
            variable = computation.variables[self.__variable_index]
            computation.remove_variable(variable)

        def _get_modified_state(self):
            computation = self.__document_model.get_computation_by_uuid(self.__computation_uuid)
            return computation.modified_state

        def _set_modified_state(self, modified_state) -> None:
            computation = self.__document_model.get_computation_by_uuid(self.__computation_uuid)
            computation.modified_state = modified_state

        def _undo(self):
            computation = self.__document_model.get_computation_by_uuid(self.__computation_uuid)
            variable = Symbolic.ComputationVariable()
            variable.read_from_dict(self.__variable_dict)
            computation.insert_variable(self.__variable_index, variable)

        def _redo(self):
            computation = self.__document_model.get_computation_by_uuid(self.__computation_uuid)
            computation.remove_variable(computation.variables[self.__variable_index])

    class CreateComputationCommand(Undo.UndoableCommand):

        def __init__(self, document_model, data_item):
            super().__init__(_("Create Computation"))
            self.__document_model = document_model
            self.__data_item_uuid = data_item.uuid
            self.__computation_uuid = None

        def close(self):
            self.__document_model = None
            self.__data_item_uuid = None
            self.__computation_uuid = None
            super().close()

        @property
        def _computation(self) -> Symbolic.Computation:
            return self.__document_model.create_computation()

        def perform(self):
            data_item = self.__document_model.get_data_item_by_uuid(self.__data_item_uuid)
            computation = self.__document_model.create_computation()
            self.__document_model.set_data_item_computation(data_item, computation)
            self.__computation_uuid = computation.uuid

        def _get_modified_state(self):
            return self.__document_model.modified_state

        def _set_modified_state(self, modified_state) -> None:
            self.__document_model.modified_state = modified_state

        def _undo(self):
            data_item = self.__document_model.get_data_item_by_uuid(self.__data_item_uuid)
            self.__document_model.set_data_item_computation(data_item, None)

    class ChangeComputationCommand(Undo.UndoableCommand):

        def __init__(self, document_model, computation: Symbolic.Computation, *, title: str=None, command_id: str=None, is_mergeable: bool=False, **kwargs):
            super().__init__(title if title else _("Change Computation"), command_id=command_id, is_mergeable=is_mergeable)
            self.__document_model = document_model
            self.__computation_uuid = computation.uuid
            self.__properties = {key: getattr(computation, key) for key in kwargs.keys()}
            self.__value_dict = kwargs
            self.initialize()

        def close(self):
            self.__properties = None
            self.__computation_uuid = None
            self.__properties = None
            self.__value_dict = None
            super().close()

        def perform(self):
            computation = self.__document_model.get_computation_by_uuid(self.__computation_uuid)
            for key, value in self.__value_dict.items():
                setattr(computation, key, value)

        def _get_modified_state(self):
            computation = self.__document_model.get_computation_by_uuid(self.__computation_uuid)
            return computation.modified_state

        def _set_modified_state(self, modified_state) -> None:
            computation = self.__document_model.get_computation_by_uuid(self.__computation_uuid)
            computation.modified_state = modified_state

        def _undo(self):
            computation = self.__document_model.get_computation_by_uuid(self.__computation_uuid)
            properties = self.__properties
            self.__properties = computation.write_to_dict()
            computation.read_from_dict(properties)

        def can_merge(self, command: Undo.UndoableCommand) -> bool:
            return isinstance(command, ComputationModel.ChangeComputationCommand) and self.command_id and self.command_id == command.command_id and self.__computation_uuid == command.__computation_uuid

    class ChangeVariableCommand(Undo.UndoableCommand):

        def __init__(self, document_model, computation: Symbolic.Computation, variable: Symbolic.ComputationVariable, *, title: str=None, command_id: str=None, is_mergeable: bool=False, **kwargs):
            super().__init__(title if title else _("Change Computation Variable"), command_id=command_id, is_mergeable=is_mergeable)
            self.__document_model = document_model
            self.__computation_uuid = computation.uuid
            self.__variable_index = computation.variables.index(variable)
            self.__property_keys = kwargs.keys()
            self.__properties = copy.deepcopy(kwargs)
            self.initialize()

        def close(self):
            self.__properties = None
            self.__computation_uuid = None
            self.__properties = None
            self.__property_keys = None
            super().close()

        def perform(self):
            computation = self.__document_model.get_computation_by_uuid(self.__computation_uuid)
            variable = computation.variables[self.__variable_index]
            properties = self.__properties
            self.__properties = {key: getattr(variable, key) for key in self.__property_keys}
            for key, value in properties.items():
                setattr(variable, key, value)

        def _get_modified_state(self):
            computation = self.__document_model.get_computation_by_uuid(self.__computation_uuid)
            return computation.modified_state

        def _set_modified_state(self, modified_state) -> None:
            computation = self.__document_model.get_computation_by_uuid(self.__computation_uuid)
            computation.modified_state = modified_state

        def _undo(self):
            computation = self.__document_model.get_computation_by_uuid(self.__computation_uuid)
            variable = computation.variables[self.__variable_index]
            properties = self.__properties
            self.__properties = {key: getattr(variable, key) for key in self.__property_keys}
            for key, value in properties.items():
                setattr(variable, key, value)

        def can_merge(self, command: Undo.UndoableCommand) -> bool:
            return isinstance(command, ComputationModel.ChangeVariableCommand) and self.command_id and self.command_id == command.command_id and self.__computation_uuid == command.__computation_uuid and self.__variable_index == command.__variable_index

    def add_variable(self, name: str=None, value_type: str=None, value=None, value_default=None, value_min=None, value_max=None, control_type: str=None, specifier: dict=None, label: str=None) -> None:
        computation = self.__computation
        if computation:
            command = ComputationModel.AddVariableCommand(self.document_controller.document_model, computation, name, value_type, value, value_default, value_min, value_max, control_type, specifier, label)
            command.perform()
            self.document_controller.push_undo_command(command)

    def remove_variable(self, variable: Symbolic.ComputationVariable) -> None:
        computation = self.__computation
        if computation:
            command = ComputationModel.RemoveVariableCommand(self.document_controller.document_model, computation, variable)
            command.perform()
            self.document_controller.push_undo_command(command)

    @property
    def computation_label(self):
        return self.__computation_label

    @computation_label.setter
    def computation_label(self, label):
        data_item = self.__display_specifier.data_item
        if data_item:
            computation = self.document_controller.document_model.get_data_item_computation(data_item)
            if not computation:
                command = ComputationModel.CreateComputationCommand(self.document_controller.document_model, data_item)
                command.perform()
                self.document_controller.push_undo_command(command)
                computation = command._computation
            command = ComputationModel.ChangeComputationCommand(self.document_controller.document_model, computation, command_id="computation_change_label", is_mergeable=True, label=label)
            command.perform()
            self.document_controller.push_undo_command(command)

    @property
    def computation_text(self):
        return self.__computation_text

    @computation_text.setter
    def computation_text(self, computation_text):
        data_item = self.__display_specifier.data_item
        if data_item:
            computation = self.document_controller.document_model.get_data_item_computation(data_item)
            if not computation:
                command = ComputationModel.CreateComputationCommand(self.document_controller.document_model, data_item)
                command.perform()
                self.document_controller.push_undo_command(command)
                computation = command._computation
            command = ComputationModel.ChangeComputationCommand(self.document_controller.document_model, computation, command_id="computation_change_label", is_mergeable=True, expression=computation_text)
            command.perform()
            self.document_controller.push_undo_command(command)

    @property
    def error_text(self):
        return self.__error_text

    def __update_computation_label(self, computation_label):
        if self.__computation_label != computation_label:
            self.__computation_label = computation_label
            self.computation_label_changed_event.fire(self.__computation_label)

    def __update_computation_text(self, computation_text):
        if self.__computation_text != computation_text:
            self.__computation_text = computation_text
            self.computation_text_changed_event.fire(self.__computation_text)

    def __update_error_text(self, error_text):
        if self.__error_text != error_text:
            self.__error_text = error_text
            self.error_text_changed_event.fire(self.__error_text)

    def clear(self):
        document_model = self.document_controller.document_model
        document_model.set_data_item_computation(self.__display_specifier.data_item, None)

    def __update_computation_display(self) -> None:
        def update_computation_display():
            label = None
            expression = None
            error_text = None
            computation = self.__computation
            if computation:
                error_text = computation.error_text
                expression = computation.expression
                label = computation.label
            self.__update_computation_label(label)
            self.__update_computation_text(expression)
            self.__update_error_text(error_text)
        self.document_controller.queue_task(update_computation_display)

    def __variable_inserted(self, index: int, variable: Symbolic.ComputationVariable) -> None:
        self.variable_inserted_event.fire(index, variable)
        def handle_property_changed(key: str) -> None:
            self.__update_computation_display()
            self.variable_property_changed_event.fire(variable)
        self.__variable_property_changed_event_listeners[variable.uuid] = variable.property_changed_event.listen(handle_property_changed)

    def __variable_removed(self, index: int, variable: Symbolic.ComputationVariable) -> None:
        self.variable_removed_event.fire(index, variable)
        self.__variable_property_changed_event_listeners[variable.uuid].close()
        del self.__variable_property_changed_event_listeners[variable.uuid]

    # not thread safe
    def __set_display_specifier(self, display_specifier):
        if self.__display_specifier != display_specifier:
            if self.__computation_changed_or_mutated_event_listener:
                self.__computation_changed_or_mutated_event_listener.close()
                self.__computation_changed_or_mutated_event_listener = None
            if self.__computation_variable_inserted_event_listener:
                self.__computation_variable_inserted_event_listener.close()
                self.__computation_variable_inserted_event_listener = None
            if self.__computation_variable_removed_event_listener:
                self.__computation_variable_removed_event_listener.close()
                self.__computation_variable_removed_event_listener = None
            if self.__computation_property_changed_event_listener:
                self.__computation_property_changed_event_listener.close()
                self.__computation_property_changed_event_listener = None
            computation = self.__computation
            if computation:
                for index, variable in enumerate(computation.variables):
                    self.__variable_removed(0, variable)
            self.__display_specifier = copy.copy(display_specifier)
            computation = self.__computation
            if computation:
                document_model = self.document_controller.document_model
                def computation_updated(data_item, computation):
                    if data_item == self.__display_specifier.data_item:
                        self.__update_computation_display()
                def property_changed(property):
                    if property == "error_text":
                        self.__update_computation_display()
                self.__computation_changed_or_mutated_event_listener = document_model.computation_updated_event.listen(computation_updated)
                self.__computation_variable_inserted_event_listener = computation.variable_inserted_event.listen(self.__variable_inserted)
                self.__computation_variable_removed_event_listener = computation.variable_removed_event.listen(self.__variable_removed)
                self.__computation_property_changed_event_listener = computation.property_changed_event.listen(property_changed)
            self.__update_computation_display()
            if computation:
                for index, variable in enumerate(computation.variables):
                    self.__variable_inserted(index, variable)


class ChangeVariableBinding(Binding.PropertyBinding):
    def __init__(self, document_controller, computation, variable, property_name: str, converter=None, fallback=None):
        super().__init__(variable, property_name, converter=converter, fallback=fallback)
        self.__property_name = property_name
        self.__old_source_setter = self.source_setter

        def set_value(value):
            if value != getattr(variable, property_name):
                command = ComputationModel.ChangeVariableCommand(document_controller.document_model, computation, self.source, **{self.__property_name: value})
                command.perform()
                document_controller.push_undo_command(command)

        self.source_setter = set_value


class ComputationPanelSection:

    def __init__(self, document_controller, computation, variable, on_remove, queue_task_fn):
        ui = document_controller.ui

        self.variable = variable

        section_widget = ui.create_column_widget()
        section_title_row = ui.create_row_widget()

        twist_down_canvas_item = CanvasItem.TwistDownCanvasItem()
        twist_down_canvas_widget = ui.create_canvas_widget(properties={"height": 20, "width": 20})
        twist_down_canvas_widget.canvas_item.add_canvas_item(twist_down_canvas_item)
        section_title_row.add(twist_down_canvas_widget)
        twist_down_label_widget = ui.create_label_widget(variable.name, properties={"stylesheet": "font-weight: bold"})
        twist_down_label_widget.bind_text(Binding.PropertyBinding(variable, "name"))
        section_title_row.add(twist_down_label_widget)
        section_title_row.add_stretch()
        section_widget.add(section_title_row)

        # boolean, integral, real, data item, region

        def make_label_row(ui, label):
            label_row = ui.create_row_widget()
            label_row.add_spacing(8)
            label_row.add(ui.create_label_widget(label))
            label_row.add_stretch()
            return label_row

        def make_name_type_row(ui, variable: Symbolic.ComputationVariable, on_change_type_fn, on_remove_fn):
            name_text_edit = ui.create_line_edit_widget()
            name_text_edit.bind_text(ChangeVariableBinding(document_controller, computation, variable, "name"))

            type_items = [("boolean", _("Boolean")), ("integral", _("Integer")), ("real", _("Real")), ("data_item", _("Data Item")), ("region", _("Region"))]
            type_combo_box = ui.create_combo_box_widget(items=type_items, item_getter=operator.itemgetter(1))

            remove_button = ui.create_push_button_widget(_("X"))

            name_type_row = ui.create_row_widget()
            name_type_row.add_spacing(8)
            name_type_row.add(name_text_edit)
            name_type_row.add_spacing(8)
            name_type_row.add(type_combo_box)
            name_type_row.add_spacing(8)
            name_type_row.add(remove_button)
            name_type_row.add_stretch()

            variable_type = variable.variable_type
            type_combo_box.current_item = next((i for i in type_items if i[0] == variable_type), None)
            type_combo_box.on_current_item_changed = lambda item: on_change_type_fn(item[0])
            remove_button.on_clicked = on_remove_fn

            return name_type_row

        def make_boolean_row(ui, variable: Symbolic.ComputationVariable, on_change_type_fn, on_remove_fn):
            name_type_row = make_name_type_row(ui, variable, on_change_type_fn, on_remove_fn)

            value_check_box = ui.create_check_box_widget(_("Value"))

            value_row = ui.create_row_widget()
            value_row.add_spacing(8)
            value_row.add(value_check_box)
            value_row.add_stretch()

            label_text_edit = ui.create_line_edit_widget()
            label_text_edit.bind_text(ChangeVariableBinding(document_controller, computation, variable, "label"))

            display_row = ui.create_row_widget()
            display_row.add_spacing(8)
            display_row.add(ui.create_label_widget(_("Label")))
            display_row.add_spacing(8)
            display_row.add(label_text_edit)
            display_row.add_stretch()

            column = ui.create_column_widget()
            column.add(make_label_row(ui, _("Variable Name / Type")))
            column.add(name_type_row)
            column.add(value_row)
            column.add(display_row)

            value_check_box.bind_checked(ChangeVariableBinding(document_controller, computation, variable, "value"))

            return column

        def make_number_row(ui, variable: Symbolic.ComputationVariable, converter, on_change_type_fn, on_remove_fn):
            name_type_row = make_name_type_row(ui, variable, on_change_type_fn, on_remove_fn)

            value_text_edit = ui.create_line_edit_widget()

            value_default_text_edit = ui.create_line_edit_widget()

            value_min_text_edit = ui.create_line_edit_widget()

            value_max_text_edit = ui.create_line_edit_widget()

            value_row = ui.create_row_widget()
            value_row.add_spacing(8)
            value_row.add(value_text_edit)
            value_row.add_spacing(4)
            value_row.add(value_default_text_edit)
            value_row.add_spacing(4)
            value_row.add(value_min_text_edit)
            value_row.add_spacing(4)
            value_row.add(value_max_text_edit)
            value_row.add_stretch()

            label_text_edit = ui.create_line_edit_widget()
            label_text_edit.bind_text(ChangeVariableBinding(document_controller, computation, variable, "label"))

            display_items = [("field", _("Field")), ("slider", _("Slider"))]
            display_combo_box = ui.create_combo_box_widget(items=display_items, item_getter=operator.itemgetter(1))

            display_row = ui.create_row_widget()
            display_row.add_spacing(8)
            display_row.add(label_text_edit)
            # display_row.add_spacing(4)
            # display_row.add(display_combo_box)
            display_row.add_stretch()

            column = ui.create_column_widget()
            column.add(make_label_row(ui, _("Variable Name / Type")))
            column.add(name_type_row)
            column.add(make_label_row(ui, _("Value / Default / Min / Max")))
            column.add(value_row)
            column.add(make_label_row(ui, _("Label / Display Type")))
            column.add(display_row)

            value_text_edit.bind_text(ChangeVariableBinding(document_controller, computation, variable, "value", converter=converter))
            value_default_text_edit.bind_text(ChangeVariableBinding(document_controller, computation, variable, "value_default", converter=converter))
            value_min_text_edit.bind_text(ChangeVariableBinding(document_controller, computation, variable, "value_min", converter=converter))
            value_max_text_edit.bind_text(ChangeVariableBinding(document_controller, computation, variable, "value_max", converter=converter))

            # display_combo_box.on_current_item_changed ... handle undo
            # display_combo_box.current_item = display_items[1]

            return column

        def make_specifier_row(ui, variable: Symbolic.ComputationVariable, on_change_type_fn, on_remove_fn, *, include_secondary=False):
            column = ui.create_column_widget()

            name_type_row = make_name_type_row(ui, variable, on_change_type_fn, on_remove_fn)

            def make_uuid_row(label, binding_identifier):
                uuid_text_edit = ui.create_line_edit_widget()
                uuid_row = ui.create_row_widget()
                uuid_row.add_spacing(8)
                uuid_row.add(ui.create_label_widget(label))
                uuid_row.add_spacing(8)
                uuid_row.add(uuid_text_edit)
                uuid_row.add_spacing(8)
                uuid_text_edit.bind_text(ChangeVariableBinding(document_controller, computation, variable, binding_identifier))
                return uuid_row

            label_text_edit = ui.create_line_edit_widget()
            label_text_edit.bind_text(ChangeVariableBinding(document_controller, computation, variable, "label"))

            display_row = ui.create_row_widget()
            display_row.add_spacing(8)
            display_row.add(ui.create_label_widget(_("Label")))
            display_row.add_spacing(8)
            display_row.add(label_text_edit)
            display_row.add_stretch()

            column.add(name_type_row)
            column.add(make_uuid_row(_("Data Item UUID"), "specifier_uuid_str"))
            if include_secondary:
                column.add(make_uuid_row(_("Region UUID"), "secondary_specifier_uuid_str"))
            column.add(display_row)

            return column

        def make_empty_row(ui, variable: Symbolic.ComputationVariable, on_change_type_fn, on_remove_fn):
            column = ui.create_column_widget()
            name_type_row = make_name_type_row(ui, variable, on_change_type_fn, on_remove_fn)
            column.add(name_type_row)
            return column

        stack = ui.create_column_widget()

        section_widget.add(stack)
        section_widget.add_spacing(4)

        def toggle():
            twist_down_canvas_item.checked = not twist_down_canvas_item.checked
            stack.visible = twist_down_canvas_item.checked
        section_open = False
        twist_down_canvas_item.checked = section_open
        stack.visible = section_open
        twist_down_canvas_item.on_button_clicked = toggle

        def change_type(variable_type):
            command = ComputationModel.ChangeVariableCommand(document_controller.document_model, computation, variable, title=_("Remove Input Data Item"), variable_type=variable_type)
            command.perform()
            document_controller.push_undo_command(command)

        def select_stack(stack, variable, specifier):
            stack.remove_all()
            variable_type = variable.variable_type
            if variable_type == "boolean":
                stack.add(make_boolean_row(ui, variable, change_type, on_remove))
            elif variable_type == "integral":
                stack.add(make_number_row(ui, variable, Converter.IntegerToStringConverter(), change_type, on_remove))
            elif variable_type == "real":
                stack.add(make_number_row(ui, variable, Converter.FloatToStringConverter(), change_type, on_remove))
            elif variable_type == "data_item":
                stack.add(make_specifier_row(ui, variable, change_type, on_remove, include_secondary=True))
            elif variable_type == "region":
                stack.add(make_specifier_row(ui, variable, change_type, on_remove))
            else:
                stack.add(make_empty_row(ui, variable, change_type, on_remove))

        def do_select_stack():
            # select stack will remove the inspector widgets, so delay it until the
            # current event (combo box changed) has finished by queueing it.
            queue_task_fn(functools.partial(select_stack, stack, variable, variable.specifier))

        self.__variable_type_changed_event_listener = variable.variable_type_changed_event.listen(do_select_stack)

        select_stack(stack, variable, variable.specifier)

        self.widget = section_widget

    def close(self):
        self.__variable_type_changed_event_listener.close()
        self.__variable_type_changed_event_listener = None


def make_image_chooser(document_controller, computation, variable, drag_fn):
    ui = document_controller.ui
    document_model = document_controller.document_model
    # drag_fn is necessary because it is unsafe to start a drag on the column containing the thumbnail
    # since dragging onto itself may delete the column during the drag.
    column = ui.create_column_widget(properties={"width": 80})
    label_row = ui.create_row_widget()
    label_widget = ui.create_label_widget(variable.display_label, properties={"width": 80})
    label_widget.bind_text(Binding.PropertyBinding(variable, "display_label"))
    label_row.add_stretch()
    label_row.add(label_widget)
    label_row.add_stretch()
    base_variable_specifier = copy.copy(variable.specifier)

    bound_data_source = document_model.resolve_object_specifier(base_variable_specifier)
    data_item = bound_data_source.value.data_item if bound_data_source else None

    def drop_mime_data(mime_data, x, y):
        data_source_mime_str = mime_data.data_as_string(DataItem.DataSource.DATA_SOURCE_MIME_TYPE)
        if data_source_mime_str:
            data_source_mime_data = json.loads(data_source_mime_str)
            data_item_uuid = uuid.UUID(data_source_mime_data["data_item_uuid"])
            data_item = document_model.get_data_item_by_key(data_item_uuid)
            variable_specifier = document_model.get_object_specifier(data_item)
            secondary_specifier = None
            if "graphic_uuid" in data_source_mime_data:
                graphic_uuid = uuid.UUID(data_source_mime_data["graphic_uuid"])
                graphic = document_model.get_graphic_by_uuid(graphic_uuid)
                if graphic:
                    secondary_specifier = document_model.get_object_specifier(graphic)
            properties = {"variable_type": "data_item", "secondary_specifier": secondary_specifier, "specifier": variable_specifier}
            command = ComputationModel.ChangeVariableCommand(document_controller.document_model, computation, variable, title=_("Remove Input Data Item"), **properties)
            command.perform()
            document_controller.push_undo_command(command)
            return "copy"
        if mime_data.has_format("text/data_item_uuid"):
            data_item_uuid = uuid.UUID(mime_data.data_as_string("text/data_item_uuid"))
            data_item = document_model.get_data_item_by_key(data_item_uuid)
            variable_specifier = document_model.get_object_specifier(data_item)
            properties = {"variable_type": "data_item", "secondary_specifier": dict(), "specifier": variable_specifier}
            command = ComputationModel.ChangeVariableCommand(document_controller.document_model, computation, variable, title=_("Remove Input Data Item"), **properties)
            command.perform()
            document_controller.push_undo_command(command)
            return "copy"
        return None

    def data_item_delete():
        variable_specifier = {"type": variable.variable_type, "version": 1, "uuid": str(uuid.uuid4())}
        command = ComputationModel.ChangeVariableCommand(document_controller.document_model, computation, variable, title=_("Remove Input Data Item"), specifier=variable_specifier)
        command.perform()
        document_controller.push_undo_command(command)

    data_item_thumbnail_source = DataItemThumbnailWidget.DataItemThumbnailSource(ui, data_item=data_item)
    data_item_chooser_widget = DataItemThumbnailWidget.ThumbnailWidget(ui, data_item_thumbnail_source, Geometry.IntSize(80, 80))

    def thumbnail_widget_drag(mime_data, thumbnail, hot_spot_x, hot_spot_y):
        # use this convoluted base object for drag so that it doesn't disappear after the drag.
        drag_fn(mime_data, thumbnail, hot_spot_x, hot_spot_y)

    data_item_chooser_widget.on_drag = thumbnail_widget_drag
    data_item_chooser_widget.on_delete = data_item_delete
    data_item_chooser_widget.on_drop_mime_data = drop_mime_data

    def property_changed(key):
        if key == "specifier":
            base_variable_specifier = copy.copy(variable.specifier)
            bound_data_item = document_model.resolve_object_specifier(base_variable_specifier)
            data_item = bound_data_item.value if bound_data_item else None
            data_item_thumbnail_source.set_data_item(data_item)

    property_changed_listener = variable.property_changed_event.listen(property_changed)
    column.add_spacing(4)
    column.add(data_item_chooser_widget)
    column.add(label_row)
    column.add_spacing(4)
    return column, [property_changed_listener]


class EditComputationDialog(Dialog.ActionDialog):

    def __init__(self, document_controller, data_item):
        ui = document_controller.ui
        super().__init__(ui, _("Edit Computation"), app=document_controller.app, parent_window=document_controller, persistent_id="EditComputationDialog" + str(data_item.uuid))

        self.ui = ui
        self.document_controller = document_controller
        self.data_item = data_item

        self._create_menus()

        self.__computation_model = ComputationModel(document_controller)

        self.__sections = list()

        label_edit_widget = ui.create_line_edit_widget(properties={"stylesheet": "min-width: 120"})
        label_edit_widget.placeholder_text = _("Computation Label")

        label_row = ui.create_row_widget()
        label_row.add_spacing(8)
        label_row.add(ui.create_label_widget(_("Description")))
        label_row.add_spacing(8)
        label_row.add(label_edit_widget)
        label_row.add_spacing(8)
        label_row.add_stretch()

        buttons_row = ui.create_row_widget()
        add_variable_button = ui.create_push_button_widget(_("Add Variable"))
        add_object_button = ui.create_push_button_widget(_("Add Object"))
        buttons_row.add(add_variable_button)
        buttons_row.add_spacing(8)
        buttons_row.add(add_object_button)
        buttons_row.add_spacing(8)
        buttons_row.add_stretch()

        # sizing in widget space (Qt) is difficult to get right and there seems to be bugs.
        # in this case, two different elements are used to effectively make a minimum window
        # size -- the text edit widget for the height and the error row for the width.
        # if both are done on the text edit widget itself, which would be preferred, Qt seems
        # to give up on layout when the scroll bar appears for too many lines.

        text_edit_row = ui.create_row_widget()
        text_edit = ui.create_text_edit_widget(properties={"stylesheet": "min-height: 180"})
        text_edit.placeholder_text = _("No Computation")
        text_edit_row.add_spacing(8)
        text_edit_row.add(text_edit)
        text_edit_row.add_spacing(8)

        error_row = ui.create_row_widget(properties={"stylesheet": "min-width: 400"})  # the stylesheet allows it to shrink. guh.
        error_label = ui.create_label_widget("\n", properties={"stylesheet": "color: red; min-width: 120"})
        error_label.word_wrap = True
        error_row.add_spacing(8)
        error_row.add(error_label)
        error_row.add_spacing(8)

        self.__text_edit = text_edit  # for testing
        self.__error_label = error_label  # for testing

        new_button = ui.create_push_button_widget(_("New"))
        update_button = ui.create_push_button_widget(_("Update"))
        button_row = ui.create_row_widget()
        button_row.add_stretch()
        button_row.add(new_button)
        button_row.add_spacing(8)
        button_row.add(update_button)
        button_row.add_spacing(8)

        self.__data_item_row = ui.create_row_widget()

        self.__variable_column = ui.create_column_widget()

        def add_object_pressed():
            document_model = document_controller.document_model
            object_specifier = document_model.get_object_specifier(document_model.data_items[0])
            self.__computation_model.add_variable("".join([random.choice(string.ascii_lowercase) for _ in range(4)]), specifier=object_specifier)

        add_object_button.on_clicked = add_object_pressed

        def add_variable_pressed():
            self.__computation_model.add_variable("".join([random.choice(string.ascii_lowercase) for _ in range(4)]), value_type="integral", value=0)

        add_variable_button.on_clicked = add_variable_pressed

        def new_pressed():
            document_controller.processing_computation(text_edit.text)

        def update_pressed():
            if text_edit.text:
                self.__computation_model.computation_text = text_edit.text
            else:
                self.__computation_model.clear()

        new_button.on_clicked = new_pressed
        update_button.on_clicked = update_pressed
        def editing_finished(text):
            if self.__computation_model:
                self.__computation_model.computation_label = text
        label_edit_widget.on_editing_finished = editing_finished

        def computation_label_changed(text):
            label_edit_widget.text = text
            if label_edit_widget.focused:
                label_edit_widget.request_refocus()

        self.__computation_label_changed_event_listener = self.__computation_model.computation_label_changed_event.listen(computation_label_changed)

        def computation_text_changed(computation_text):
            text_edit.text = computation_text

        self.__computation_text_changed_event_listener = self.__computation_model.computation_text_changed_event.listen(computation_text_changed)

        def error_text_changed(error_text):
            error_label.text = error_text

        self.__error_text_changed_event_listener = self.__computation_model.error_text_changed_event.listen(error_text_changed)

        self.__listeners = list()

        def rebuild_data_item_row():
            self.__data_item_row.remove_all()
            for listener in self.__listeners:
                listener.close()
            self.__listeners = list()
            self.__data_item_row.add_spacing(8)
            for section in self.__sections:
                variable = section.variable
                if variable.variable_type in ("data_item", ):
                    widget, listeners = make_image_chooser(document_controller, self.__computation_model.computation, variable, self.content.drag)
                    self.__listeners.extend(listeners)
                    self.__data_item_row.add(widget)
                    self.__data_item_row.add_spacing(8)
            self.__data_item_row.add_stretch()
            target_column = ui.create_column_widget(properties={"width": 80})

            def thumbnail_widget_drag(mime_data, thumbnail, hot_spot_x, hot_spot_y):
                # use this convoluted base object for drag so that it doesn't disappear after the drag.
                self.content.drag(mime_data, thumbnail, hot_spot_x, hot_spot_y)

            data_item_thumbnail_source = DataItemThumbnailWidget.DataItemThumbnailSource(ui, data_item=data_item)  # TODO: never closed
            data_item_chooser_widget = DataItemThumbnailWidget.ThumbnailWidget(ui, data_item_thumbnail_source, Geometry.IntSize(80, 80))
            data_item_chooser_widget.on_drag = thumbnail_widget_drag
            target_column.add_spacing(4)
            target_column.add(data_item_chooser_widget)
            target_column.add(ui.create_label_widget(_("Target"), properties={"width": 80}))
            target_column.add_spacing(4)
            self.__data_item_row.add(target_column)
            self.__data_item_row.add_spacing(8)

        def variable_inserted(index: int, variable: Symbolic.ComputationVariable) -> None:
            def remove_variable():
                self.__computation_model.remove_variable(variable)
            section = ComputationPanelSection(document_controller, self.__computation_model.computation, variable, remove_variable, self.document_controller.queue_task)
            self.__variable_column.insert(section.widget, index)
            self.__sections.insert(index, section)
            rebuild_data_item_row()

        def variable_removed(index: int, variable: Symbolic.ComputationVariable) -> None:
            self.__variable_column.remove(self.__variable_column.children[index])
            if self.__sections[index]:
                self.__sections[index].close()
            del self.__sections[index]
            rebuild_data_item_row()

        def variable_property_changed(variable: Symbolic.ComputationVariable) -> None:
            rebuild_data_item_row()

        self.__variable_inserted_event_listener = self.__computation_model.variable_inserted_event.listen(variable_inserted)
        self.__variable_removed_event_listener = self.__computation_model.variable_removed_event.listen(variable_removed)
        self.__variable_property_changed_event_listener = self.__computation_model.variable_property_changed_event.listen(variable_property_changed)

        # for testing
        self._new_button = new_button
        self._update_button = update_button

        self.__computation_model.set_data_item(data_item)

        column = self.content
        column.add_spacing(6)
        column.add(label_row)
        column.add(self.__data_item_row)
        column.add(self.__variable_column)
        column.add(buttons_row)
        column.add_spacing(6)
        column.add(text_edit_row)
        column.add_spacing(6)
        column.add(error_row)
        column.add_spacing(6)
        column.add(button_row)
        column.add_spacing(6)

    def close(self):
        self.document_controller.clear_task(str(id(self)))
        for listener in self.__listeners:
            listener.close()
        self.__listeners = list()
        self.__computation_label_changed_event_listener.close()
        self.__computation_label_changed_event_listener = None
        self.__computation_text_changed_event_listener.close()
        self.__computation_text_changed_event_listener = None
        self.__error_text_changed_event_listener.close()
        self.__error_text_changed_event_listener = None
        self.__variable_inserted_event_listener.close()
        self.__variable_inserted_event_listener = None
        self.__variable_removed_event_listener.close()
        self.__variable_removed_event_listener = None
        self.__computation_model.close()
        self.__computation_model = None
        super().close()

    def size_changed(self, width, height):
        self.__error_label.size = Geometry.IntSize(height=self.__error_label.size.height, width=self.__text_edit.size.width)

    @property
    def _sections_for_testing(self):
        return self.__sections

    @property
    def _variable_column_for_testing(self):
        return self.__variable_column

    @property
    def _text_edit_for_testing(self):
        return self.__text_edit

    @property
    def _error_label_for_testing(self):
        return self.__error_label

    @property
    def _computation_model_for_testing(self):
        return self.__computation_model



"""
    Contains classes related to connections between library objects.
"""

# standard libraries
import copy
import functools
import weakref

# third party libraries
# None

# local libraries
from nion.swift.model import Graphics
from nion.utils import Binding
from nion.utils import Converter
from nion.utils import Event
from nion.utils import Observable
from nion.utils import Persistence


class Connection(Observable.Observable, Persistence.PersistentObject):
    """ Represents a connection between two objects. """

    def __init__(self, type, *, parent=None):
        super().__init__()
        self.__container_weak_ref = None
        self.about_to_be_removed_event = Event.Event()
        self._about_to_be_removed = False
        self._closed = False
        self.__registration_listener = None
        self.define_type(type)
        self.define_property("parent_uuid", converter=Converter.UuidToStringConverter())
        self.__parent = parent
        if parent is not None:
            self.parent_uuid = parent.uuid

    def close(self):
        if self.__registration_listener:
            self.__registration_listener.close()
            self.__registration_listener = None
        assert self._about_to_be_removed
        assert not self._closed
        self._closed = True
        self.__container_weak_ref = None

    @property
    def container(self):
        return self.__container_weak_ref()

    def about_to_be_inserted(self, container):
        assert self.__container_weak_ref is None
        self.__container_weak_ref = weakref.ref(container)

    def about_to_be_removed(self):
        # called before close and before item is removed from its container
        self.about_to_be_removed_event.fire()
        assert not self._about_to_be_removed
        self._about_to_be_removed = True
        self.__container_weak_ref = None

    def clone(self) -> "Connection":
        connection = copy.deepcopy(self)
        connection.uuid = self.uuid
        return connection

    def _property_changed(self, name, value):
        self.notify_property_changed(name)

    @property
    def parent(self):
        return self.__parent

    @parent.setter
    def parent(self, parent):
        self.__parent = parent
        self.parent_uuid = parent.uuid if parent else None

    def persistent_object_context_changed(self):
        """ Override from PersistentObject. """
        super().persistent_object_context_changed()

        def parent_registered(parent):
            self.__parent = parent

        def parent_unregistered(parent=None):
            pass

        def change_registration(registered_object, unregistered_object):
            if registered_object and registered_object.uuid == self.parent_uuid:
                self.__parent = registered_object

        if self.persistent_object_context:
            self.__registration_listener = self.persistent_object_context.registration_event.listen(change_registration)

            self.__parent = self.persistent_object_context.get_registered_object(self.parent_uuid)

            # self.persistent_object_context.subscribe(self.parent_uuid, parent_registered, parent_unregistered)
        else:
            parent_unregistered()


class PropertyConnection(Connection):
    """ Binds the properties of two objects together. """

    def __init__(self, source=None, source_property=None, target=None, target_property=None, *, parent=None):
        super().__init__("property-connection", parent=parent)
        self.define_property("source_uuid", converter=Converter.UuidToStringConverter())
        self.define_property("source_property")
        self.define_property("target_uuid", converter=Converter.UuidToStringConverter())
        self.define_property("target_property")
        # these are only set in persistent object context changed
        self.__source = None
        self.__target = None
        self.__binding = None
        self.__target_property_changed_listener = None
        self.__registration_listener = None
        # suppress messages while we're setting source or target
        self.__suppress = False
        # but setup if we were passed objects
        if source is not None:
            self.source_uuid = source.uuid
        if source_property:
            self.source_property = source_property
        if target is not None:
            self.target_uuid = target.uuid
        if target_property:
            self.target_property = target_property

    def close(self):
        if self.__registration_listener:
            self.__registration_listener.close()
            self.__registration_listener = None
        if self.__binding:
            self.__binding.close()
            self.__binding = None
        if self.__target_property_changed_listener:
            self.__target_property_changed_listener.close()
            self.__target_property_changed_listener = None
        super().close()

    @property
    def _source(self):
        return self.__source

    @property
    def _target(self):
        return self.__target

    def __set_target_from_source(self, value):
        assert not self._closed
        if not self.__suppress:
            self.__suppress = True
            setattr(self.__target, self.target_property, value)
            self.__suppress = False

    def __set_source_from_target(self, value):
        assert not self._closed
        if not self.__suppress:
            self.__suppress = True
            if self.__binding:
                self.__binding.update_source(value)
            self.__suppress = False

    def persistent_object_context_changed(self):
        """ Override from PersistentObject. """
        super().persistent_object_context_changed()

        def register():
            if self.__source is not None and self.__target is not None:
                assert not self.__binding
                self.__binding = Binding.PropertyBinding(self.__source, self.source_property)
                self.__binding.target_setter = self.__set_target_from_source
                self.__binding.update_target_direct(self.__binding.get_target_value())

        def source_registered(source):
            self.__source = source
            register()

        def target_registered(target):
            self.__target = target

            def property_changed(target, property_name):
                if property_name == self.target_property:
                    self.__set_source_from_target(getattr(target, property_name))

            assert self.__target_property_changed_listener is None
            self.__target_property_changed_listener = target.property_changed_event.listen(functools.partial(property_changed, target))
            register()

        def unregistered(source=None):
            if self.__binding:
                self.__binding.close()
                self.__binding = None
            if self.__target_property_changed_listener:
                self.__target_property_changed_listener.close()
                self.__target_property_changed_listener = None

        def change_registration(registered_object, unregistered_object):
            if registered_object and registered_object.uuid == self.source_uuid:
                source_registered(registered_object)
            if registered_object and registered_object.uuid == self.target_uuid:
                target_registered(registered_object)
            if unregistered_object and unregistered_object in (self._source, self._target):
                unregistered(unregistered_object)

        if self.persistent_object_context:
            self.__registration_listener = self.persistent_object_context.registration_event.listen(change_registration)
            source = self.persistent_object_context.get_registered_object(self.source_uuid)
            target = self.persistent_object_context.get_registered_object(self.target_uuid)
            if source:
                source_registered(source)
            if target:
                target_registered(target)
        else:
            unregistered()


class IntervalListConnection(Connection):
    """Binds the intervals on a display to the interval_descriptors on a line profile graphic.

    This is a one way connection from the display to the line profile graphic.
    """

    def __init__(self, display=None, line_profile=None, *, parent=None):
        super().__init__("interval-list-connection", parent=parent)
        self.define_property("source_uuid", converter=Converter.UuidToStringConverter())
        self.define_property("target_uuid", converter=Converter.UuidToStringConverter())
        # these are only set in persistent object context changed
        self.__source = display
        self.__target = line_profile
        self.__item_inserted_event_listener = None
        self.__item_removed_event_listener = None
        self.__interval_mutated_listeners = list()
        # but setup if we were passed objects
        if display is not None:
            self.source_uuid = display.uuid
        if line_profile is not None:
            self.target_uuid = line_profile.uuid

    def close(self):
        super().close()

    def persistent_object_context_changed(self):
        """ Override from PersistentObject. """
        super().persistent_object_context_changed()

        def detach():
            for listener in self.__interval_mutated_listeners:
                listener.close()
            self.__interval_mutated_listeners = list()

        def reattach():
            detach()
            interval_descriptors = list()
            if self.__source:
                for region in self.__source.graphics:
                    if isinstance(region, Graphics.IntervalGraphic):
                        interval_descriptor = {"interval": region.interval, "color": "#F00"}
                        interval_descriptors.append(interval_descriptor)
                        self.__interval_mutated_listeners.append(region.property_changed_event.listen(lambda k: reattach()))
            if self.__target:
                self.__target.interval_descriptors = interval_descriptors

        def item_inserted(key, value, before_index):
            if key == "graphics" and self.__target:
                reattach()

        def item_removed(key, value, index):
            if key == "graphics" and self.__target:
                reattach()

        def source_registered(source):
            self.__source = source
            self.__item_inserted_event_listener = self.__source.item_inserted_event.listen(item_inserted)
            self.__item_removed_event_listener = self.__source.item_removed_event.listen(item_removed)
            reattach()

        def target_registered(target):
            self.__target = target
            reattach()

        def unregistered(source=None):
            if self.__item_inserted_event_listener:
                self.__item_inserted_event_listener.close()
                self.__item_inserted_event_listener = None
            if self.__item_removed_event_listener:
                self.__item_removed_event_listener.close()
                self.__item_removed_event_listener = None

        if self.persistent_object_context:
            self.persistent_object_context.subscribe(self.source_uuid, source_registered, unregistered)
            self.persistent_object_context.subscribe(self.target_uuid, target_registered, unregistered)
        else:
            unregistered()


def connection_factory(lookup_id):
    build_map = {
        "property-connection": PropertyConnection,
        "interval-list-connection": IntervalListConnection,
    }
    type = lookup_id("type")
    return build_map[type]() if type in build_map else None
